import json

from ratio1 import Logger, const
from ratio1.bc import DefaultBlockEngine



if __name__ == '__main__' :
  l = Logger("ENC", base_folder=".", app_folder="_local_cache")
  
  
  data = {
  "EE_IS_ENCRYPTED": True,
  "EE_ENCRYPTED_DATA": "fwcctxETcji64z/KAYFmyWE8RDrJwm2BToMMZD9uesGtWatmDA1nadqpaAwlQajJg9SYb6jeadwxeIqg15VlDmNO9X2dxSiu151Dos900kx2Y8dM3q/S7A==",
  "EE_ID": "0xai_A9gulGoLPxo7hOo05GIeSvfld_IF_CucTnP00CznjmJN",
  "SESSION_ID": "SolisClient_bf2d",
  "INITIATOR_ID": "SolisClient_bf2d",
  "EE_SENDER": "0xai_Ag28FO6Yg254w-RHsBtODc_WLq7jz4SwfUAW7S7zP0U1",
  "TIME": "2025-01-07 00:37:16.291005",
  "EE_SIGN": "MEQCIH9T4_zB4v8cgjhUAwOvELrthGCUZ4v87LR-5DnFetkuAiAm2nOoZIitma10skQbuf12Yp9l_M7zyAcG9mWScuPZGw==",
  "EE_HASH": "7d7d49f39d1f870af74bece27262d9b050b5358a15c71b249150d04a712ce72c"
  }
  
  
  eng1 = DefaultBlockEngine(
    log=l, name="test1", 
    config={
        "PEM_FILE"     : "stg-0.pem",
        "PASSWORD"     : None,      
        "PEM_LOCATION" : "data"
      }
  )
  
  
  ver = eng1.verify(data)
  
  eng1.P(f"Signature: {'OK' if ver.valid else 'ERROR'}, {ver}", color='g' if ver.valid else 'r')
  
  decr = eng1.decrypt(
    encrypted_data_b64=data["EE_ENCRYPTED_DATA"],
    sender_address=data["EE_SENDER"],
    debug=True,
  )
  eng1.P(f"Decrypted: {decr}")



# BC Consts:
from ..bc.base import BCct as BC_CT

TLBR_POS = 'TLBR_POS'
PROB_PRC = 'PROB_PRC'
TYPE = 'TYPE'
COLOR_TAG = 'COLOR_TAG'

NOTIFICATION_TYPE = 'NOTIFICATION_TYPE'
STATUS_TYPE_KEY = NOTIFICATION_TYPE

# Notification types


class STATUS_TYPE:
  NOTIFICATION_TYPE = NOTIFICATION_TYPE
  STATUS_NORMAL = 'NORMAL'
  STATUS_EXCEPTION = 'EXCEPTION'
  STATUS_EMAIL = 'EMAIL'
  STATUS_ABNORMAL_FUNCTIONING = 'ABNORMAL FUNCTIONING'


class NOTIFICATION_CODES:
  # pipelines from 1 to 99
  PIPELINE_OK = 1  # basic config is ok
  PIPELINE_FAILED = -PIPELINE_OK

  PIPELINE_DATA_OK = PIPELINE_OK + 1  # data seems to flow
  PIPELINE_DATA_FAILED = -PIPELINE_DATA_OK

  PIPELINE_DCT_CONFIG_OK = PIPELINE_DATA_OK + 1  # dct config is ok
  PIPELINE_DCT_CONFIG_FAILED = -PIPELINE_DCT_CONFIG_OK

  PIPELINE_ARCHIVE_OK = PIPELINE_DCT_CONFIG_OK + 1  # archiving is ok
  PIPELINE_ARCHIVE_FAILED = -PIPELINE_ARCHIVE_OK

  # plugins from 100 to 999
  PLUGIN_CONFIG_OK = 100
  PLUGIN_CONFIG_FAILED = -PLUGIN_CONFIG_OK

  PLUGIN_INSTANCE_COMMAND_OK = 101
  PLUGIN_INSTANCE_COMMAND_FAILED = -PLUGIN_INSTANCE_COMMAND_OK

  PLUGIN_PAUSE_OK = 110
  PLUGIN_PAUSE_FAILED = -PLUGIN_PAUSE_OK

  PLUGIN_RESUME_OK = 111
  PLUGIN_RESUME_FAILED = -PLUGIN_RESUME_OK

  PLUGIN_WORKING_HOURS_SHIFT_START = 112
  PLUGIN_WORKING_HOURS_SHIFT_START_FAILED = -PLUGIN_WORKING_HOURS_SHIFT_START

  PLUGIN_WORKING_HOURS_SHIFT_END = 113
  PLUGIN_WORKING_HOURS_SHIFT_END_FAILED = -PLUGIN_WORKING_HOURS_SHIFT_END

  PLUGIN_CONFIG_IN_PAUSE_OK = 120
  PLUGIN_CONFIG_IN_PAUSE_FAILED = -PLUGIN_CONFIG_IN_PAUSE_OK

  # BUSINESS NOTIFICATIONS FROM 500 to 899

  # Example:
  # PLUGIN_SOMETHING_OK = 610
  # PLUGIN_SOMETHING_FAILED = -PLUGIN_SOMETHING_OK

  # END BUSINESS NOTIFICATIONS FROM 500 to 899

  PLUGIN_DELAYED = -999

  # comms from 1000
  COMMS_OK = 1000
  
  COMM_RECEIVED_BAD_COMMAND = -5000

  # serving from 10000
  SERVING_START_OK = 10000
  SERVING_START_FAILED = -SERVING_START_OK
  SERVING_START_FATAL_FAIL = -99999
  
  
  

  TAGS = {
    PIPELINE_OK: "PIPELINE_OK",
    PIPELINE_FAILED: "PIPELINE_FAILED",
    PIPELINE_DATA_OK: "PIPELINE_DATA_OK",
    PIPELINE_DATA_FAILED: "PIPELINE_DATA_FAILED",
    PIPELINE_DCT_CONFIG_OK: "PIPELINE_DCT_CONFIG_OK",
    PIPELINE_DCT_CONFIG_FAILED: "PIPELINE_DCT_CONFIG_FAILED",
    PIPELINE_ARCHIVE_OK: "PIPELINE_ARCHIVE_OK",
    PIPELINE_ARCHIVE_FAILED: "PIPELINE_ARCHIVE_FAILED",

    PLUGIN_CONFIG_OK: "PLUGIN_CONFIG_OK",
    PLUGIN_CONFIG_FAILED: "PLUGIN_CONFIG_FAILED",
    PLUGIN_INSTANCE_COMMAND_OK: "PLUGIN_INSTANCE_COMMAND_OK",
    PLUGIN_INSTANCE_COMMAND_FAILED: "PLUGIN_INSTANCE_COMMAND_FAILED",
    PLUGIN_PAUSE_OK: "PLUGIN_PAUSE_OK",
    PLUGIN_PAUSE_FAILED: "PLUGIN_PAUSE_FAILED",
    PLUGIN_RESUME_OK: "PLUGIN_RESUME_OK",
    PLUGIN_RESUME_FAILED: "PLUGIN_RESUME_FAILED",
    PLUGIN_DELAYED: "PLUGIN_DELAYED",
    PLUGIN_WORKING_HOURS_SHIFT_START: "PLUGIN_WORKING_HOURS_SHIFT_START",
    PLUGIN_WORKING_HOURS_SHIFT_START_FAILED: "PLUGIN_WORKING_HOURS_SHIFT_START_FAILED",
    PLUGIN_WORKING_HOURS_SHIFT_END: "PLUGIN_WORKING_HOURS_SHIFT_END",
    PLUGIN_WORKING_HOURS_SHIFT_END_FAILED: "PLUGIN_WORKING_HOURS_SHIFT_END_FAILED",

    PLUGIN_CONFIG_IN_PAUSE_OK: "PLUGIN_CONFIG_IN_PAUSE_OK",
    PLUGIN_CONFIG_IN_PAUSE_FAILED: "PLUGIN_CONFIG_IN_PAUSE_FAILED",

    SERVING_START_OK: "SERVING_START_OK",
    SERVING_START_FAILED: "SERVING_START_FAILED",
    SERVING_START_FATAL_FAIL: "SERVING_START_FATAL_FAIL",
    
    COMM_RECEIVED_BAD_COMMAND: "COMM_RECEIVED_BAD_COMMAND",

  }
  CODES = {v: k for k, v in TAGS.items()}

  # next section could be missing
  PIPELINE_OK_TAG = TAGS[PIPELINE_OK]
  PIPELINE_FAILED_TAG = TAGS[PIPELINE_FAILED]

  PLUGIN_CONFIG_OK_TAG = TAGS[PLUGIN_CONFIG_OK]
  PLUGIN_CONFIG_FAILED_TAG = TAGS[PLUGIN_CONFIG_FAILED]


class COMMANDS:
  COMMANDS = 'COMMANDS'
  RESTART = 'RESTART'
  STATUS = 'STATUS'
  STOP = 'STOP'
  UPDATE_CONFIG = 'UPDATE_CONFIG'
  DELETE_CONFIG = 'DELETE_CONFIG'
  UPDATE_PIPELINE_INSTANCE = 'UPDATE_PIPELINE_INSTANCE'
  BATCH_UPDATE_PIPELINE_INSTANCE = 'BATCH_UPDATE_PIPELINE_INSTANCE'
  PIPELINE_COMMAND = 'PIPELINE_COMMAND'
  ARCHIVE_CONFIG = 'ARCHIVE_CONFIG'
  DELETE_CONFIG_ALL = 'DELETE_CONFIG_ALL'
  ARCHIVE_CONFIG_ALL = 'ARCHIVE_CONFIG_ALL'
  ACTIVE_PLUGINS = 'ACTIVE_PLUGINS'
  RELOAD_CONFIG_FROM_DISK = 'RELOAD_CONFIG_FROM_DISK'
  FULL_HEARTBEAT = 'FULL_HEARTBEAT'
  TIMERS_ONLY_HEARTBEAT = 'TIMERS_ONLY_HEARTBEAT'
  SIMPLE_HEARTBEAT = 'SIMPLE_HEARTBEAT'
  INSTANCE_COMMAND = 'INSTANCE_COMMAND'
  COMMAND_PARAMS = 'COMMAND_PARAMS'

  FINISH_ACQUISITION = 'FINISH_ACQUISITION'


class PAYLOAD_DATA:
  EE_ENCRYPTED_DATA = 'EE_ENCRYPTED_DATA'
  EE_IS_ENCRYPTED = 'EE_IS_ENCRYPTED'
  INITIATOR_ID = 'INITIATOR_ID'
  INITIATOR_ADDR = 'INITIATOR_ADDR'
  MODIFIED_BY_ID = 'MODIFIED_BY_ID'
  MODIFIED_BY_ADDR = 'MODIFIED_BY_ADDR'
  SESSION_ID = 'SESSION_ID'
  STREAM_NAME = 'STREAM_NAME'
  NAME = 'NAME'
  INSTANCE_CONFIG = 'INSTANCE_CONFIG'
  SIGNATURE = 'SIGNATURE'
  INSTANCE_ID = 'INSTANCE_ID'
  TIME = 'TIME'
  EE_TIMESTAMP = 'EE_TIMESTAMP'
  EE_TIMEZONE = 'EE_TIMEZONE'
  EE_TZ = 'EE_TZ'
  SB_TIMESTAMP = EE_TIMESTAMP
  EE_MESSAGE_ID = 'EE_MESSAGE_ID'
  EE_MESSAGE_SEQ = 'EE_MESSAGE_SEQ'
  SB_MESSAGE_ID = EE_MESSAGE_ID
  EE_TOTAL_MESSAGES = 'EE_TOTAL_MESSAGES'
  SB_TOTAL_MESSAGES = EE_TOTAL_MESSAGES
  EE_FORMATTER = 'EE_FORMATTER'
  SB_IMPLEMENTATION = 'SB_IMPLEMENTATION'
  EE_EVENT_TYPE = 'EE_EVENT_TYPE'
  SB_EVENT_TYPE = 'SB_EVENT_TYPE'
  EE_PAYLOAD_PATH = 'EE_PAYLOAD_PATH'
  EE_PAYLOAD_INFO = 'EE_PAYLOAD_INFO'
  EE_VERSION = 'EE_VERSION'
  EE_ID = 'EE_ID'
  EE_PIPELINE_NAME = 'EE_PIPELINE_NAME'
  EE_SENDER = BC_CT.SENDER
  EE_HASH = BC_CT.HASH
  EE_SIGN = BC_CT.SIGN
  
  EE_ETH_ADDR = 'EE_ETH_SENDER'
  EE_ETH_SIGN = 'EE_ETH_SIGN'
  
  EE_DESTINATION = "EE_DEST" # can be either single address or list of addresses
  EE_DESTINATION_ID = "EE_DEST_ID" # can be either single alias or list of aliases

  NOTIFICATION = 'NOTIFICATION'
  INFO = 'INFO'

  TAGS = 'TAGS'

  ID_TAGS = 'ID_TAGS'
  
  NETMON_CURRENT_NETWORK = 'CURRENT_NETWORK'
  NETMON_STATUS_KEY = "working"
  NETMON_STATUS_ONLINE = "ONLINE"
  NETMON_ADDRESS = "address"
  NETMON_ETH_ADDRESS = "eth_address"
  NETMON_EEID = "eeid"
  NETMON_LAST_REMOTE_TIME = 'last_remote_time'
  NETMON_UPTIME = 'uptime'
  NETMON_NODE_UTC = 'node_utc'
  NETMON_LAST_SEEN = 'last_seen_sec'
  NETMON_IS_SUPERVISOR = 'is_supervisor'
  NETMON_WHITELIST = 'whitelist'
  NETMON_NODE_SECURED = 'secured'
  NETMON_NODE_VERSION = 'version'
  NETMON_NODE_R1FS_ID = 'r1fs_id'
  NETMON_NODE_R1FS_ONLINE = 'r1fs_online'
  NETMON_NODE_R1FS_RELAY = 'r1fs_relay'
  

  
  
class NET_CONFIG:
  STORE_COMMAND = "SET_CONFIG"
  REQUEST_COMMAND = "GET_CONFIG"
  NET_CONFIG_DATA = 'NET_CONFIG_DATA'
  OPERATION = 'OP'
  DESTINATION = 'DEST'
  DATA = 'DATA'
  
  PIPELINES = "PIPELINES"
  PLUGINS_STATUSES = "PLUGIN_STATUSES"
    


# TODO: for custom plugin, do the plugin verification locally too
import os
from time import sleep, time

from ..code_cheker.base import BaseCodeChecker
from ..const import PAYLOAD_DATA
from .distributed_custom_code_presets import DistributedCustomCodePresets
from .instance import Instance
from .responses import PipelineArchiveResponse, PipelineOKResponse
from .transaction import Transaction


class Pipeline(BaseCodeChecker):
  """
  A `Pipeline` is a an object that encapsulates a one-to-many, data acquisition to data processing, flow of data.

  A `Pipeline` contains one thread of data acquisition (which does not mean only one source of data), and many
  processing units, usually named `Plugins`. 

  An `Instance` is a running thread of a `Plugin` type, and one may want to have multiple `Instances`, because each can be configured independently.

  As such, one will work with `Instances`, by reffering to them with the unique identifier (Pipeline, Plugin, Instance).

  In the documentation, the following reffer to the same thing:
    `Pipeline` == `Stream`

    `Plugin` == `Signature`
  """

  def __init__(
    self, 
    session, 
    log, 
    *, 
    node_addr, 
    name, 
    config={}, 
    plugins=[], 
    on_data=None, 
    on_notification=None, 
    is_attached=False, 
    existing_config=None, 
    plugins_statuses=None,
    debug=False,
    **kwargs
  ) -> None:
    """
    A `Pipeline` is a an object that encapsulates a one-to-many, data acquisition to data processing, flow of data.

    A `Pipeline` contains one thread of data acquisition (which does not mean only one source of data), and many
    processing units, usually named `Plugins`. 

    An `Instance` is a running thread of a `Plugin` type, and one may want to have multiple `Instances`, because each can be configured independently.

    As such, one will work with `Instances`, by referring to them with the unique identifier (Pipeline, Plugin, Instance).

    In the documentation, the following refer to the same thing:
      `Pipeline` == `Stream`

      `Plugin` == `Signature`

    Parameters
    ----------
    session : Session
        The Session object which owns this pipeline. A pipeline must be attached to a Session because that is the only
        way the `on_X` callbacks are called
    log : Logger
        A logger object which implements basic logging functionality and some other utils stuff. Can be ignored for now.
        In the future, the documentation for the Logger base class will be available and developers will be able to use
        custom-made Loggers. 
    node_addr : str
        Address of the ratio1 Edge Protocol edge node that will handle this pipeline.  
    name : str
        The name of this pipeline.
    data_source : str
        This is the name of the DCT plugin, which resembles the desired functionality of the acquisition.
    config : dict, optional
        This is the dictionary that contains the configuration of the acquisition source, by default {}
    plugins : List | None, optional
        This is the list with manually configured business plugins that will be in the pipeline at creation time.
        We recommend to leave this as `[]` or as `None` and use the API to create plugin instances. 
    on_data : Callable[[Pipeline, str, str, dict], None], optional,
        Callback that handles messages received from any plugin instance. 
        As arguments, it has a reference to this Pipeline object, along with the payload itself.
        This callback acts as a default payload processor and will be called even if for a given instance
        the user has defined a specific callback.
        Defaults to None.
    on_notification : Callable[[Pipeline, dict], None], optional
        Callback that handles notifications received from this instance. 
        As arguments, it has a reference to this Pipeline object, along with the payload itself. 
        This callback acts as a default payload processor and will be called even if for a given instance
        the user has defined a specific callback.
        Defaults to None.
    is_attached : bool
        This is used internally to allow the user to create or attach to a pipeline, and then use the same
        objects in the same way, by default True
    **kwargs : dict
        The user can provide the configuration of the acquisition source directly as kwargs.
    """
    self.log = log
    self.session = session
    self.node_addr = node_addr
    self.name = name
    self.__debug = debug

    self.config = {}
    plugins = config.pop('PLUGINS', plugins)

    if is_attached:
      assert existing_config is not None, "When attaching to a pipeline, the existing configuration should be found in the heartbeat of the ratio1 Edge Protocol edge node."
      assert config == {}, "Cannot provide a configuration when attaching to a pipeline."
      assert len(kwargs) == 0, "Cannot provide a configuration when attaching to a pipeline."
      self.config = {k.upper(): v for k, v in existing_config.items()}
      self.config = self.__pop_ignored_keys_from_config(self.config)
      self.proposed_config = None
    else:
      self.proposed_config = {**config, **kwargs}
      self.proposed_config = {k.upper(): v for k, v in self.proposed_config.items()}
      self.proposed_config = self.__pop_ignored_keys_from_config(self.proposed_config)
    self.__staged_config = None


    self.__update_plugins_statuses_data(plugins_statuses)    

    self.__was_last_operation_successful = None

    self.proposed_remove_instances = []
    self.__staged_remove_instances = []

    self.on_data_callbacks = []
    self.on_notification_callbacks = []

    if on_data is not None:
      if isinstance(on_data, list):
        for cb in on_data:
          if cb is not None:
            self.on_data_callbacks.append(cb)
      elif callable(on_data):
        self.on_data_callbacks.append(on_data)
      else:
        raise ValueError("on_data should be a callable or a list of callables")
      
    if on_notification is not None:
      if isinstance(on_notification, list):
        for cb in on_data:
          if cb is not None:
            self.on_notification_callbacks.append(cb)
      elif callable(on_notification):
        self.on_notification_callbacks.append(on_notification)
      else:
        raise ValueError("on_notification should be a callable or a list of callables")

    self.lst_plugin_instances: list[Instance] = []

    self.__init_plugins(plugins, is_attached)
    return
  
  def Pd(self, *args, **kwargs):
    """
    Print debug message.
    """
    if self.__debug:
      kwargs["color"] = 'd' if kwargs.get("color") != 'r' else 'r'
      self.P(*args, **kwargs)
    return

  # Utils
  if True:
    def __init_instance(
      self, 
      signature, 
      instance_id, 
      config, 
      on_data, 
      on_notification, 
      is_attached,
      debug=False,
    ):
      """
      This method is used to create a new instance of a plugin and add it to the pipeline.
              
      """
      instance_class = None
      str_signature = None
      if isinstance(signature, str):
        instance_class = Instance
        str_signature = signature.upper()
      else:
        instance_class = signature
        str_signature = instance_class.signature.upper()
      instance = instance_class(
        self.log,
        pipeline=self,
        signature=str_signature,
        instance_id=instance_id,
        config=config,
        on_data=on_data,
        on_notification=on_notification,
        is_attached=is_attached,
        debug=debug, 
      )
      self.lst_plugin_instances.append(instance)
      return instance

    def __init_plugins(self, plugins, is_attached):
      """
      Initialize the plugins list. This method is called at the creation of the pipeline and is used to create the instances of the plugins that are part of the pipeline.

      Parameters
      ----------
      plugins : List | None
        The list of plugins, as they are found in the pipeline configuration dictionary in the heartbeat.
        
      is_attached : bool
        This is used internally to allow the user to create or attach to a pipeline, and then use the same objects in the same way.
        

      """
      if plugins is None:
        return

      for dct_signature_instances in plugins:
        signature = dct_signature_instances['SIGNATURE']
        instances = dct_signature_instances['INSTANCES']
        for dct_instance in instances:
          config = {k.upper(): v for k, v in dct_instance.items()}
          instance_id = config.pop('INSTANCE_ID')
          instance_object = self.__init_instance(signature, instance_id, config, None, None, is_attached=is_attached)
          # now we update the status of the plugin instance with the last known status
          self._update_plugin_status(instance_object)
        # end for dct_instance
      # end for dct_signature_instances
      return
    
    def __update_plugins_statuses_data(self, plugins_statuses):
      if plugins_statuses is not None:
        self.last_plugins_statuses = plugins_statuses
        self.last_plugins_statuses_time = time()
      else:
        self.last_plugins_statuses = None
        self.last_plugins_statuses_time = 0
      return


    def __get_recent_plugin_instance_status(self, signature, instance_id):
      """
      Get the most recent status of a plugin instance.
      """
      result = None
      if self.last_plugins_statuses is not None:
        for plugin_status in self.last_plugins_statuses:
          if (
            plugin_status['SIGNATURE'] == signature and 
            plugin_status['INSTANCE_ID'] == instance_id and
            plugin_status['STREAM_ID'] == self.name
          ):
            result = plugin_status
            break
      return result
    

    def _update_plugin_status(self, instance_object : Instance):
      """
      Update the status of a plugin instance using any existing last known status.
      
      """
      signature = instance_object.signature
      instance_id = instance_object.instance_id
      plugin_status = self.__get_recent_plugin_instance_status(signature, instance_id)
      if plugin_status is not None:
        instance_object.last_known_status = plugin_status
      return


    def __get_proposed_pipeline_config(self):
      """
      Construct the proposed pipeline configuration dictionary.

      Returns
      -------
      dict
          The proposed pipeline configuration dictionary.
      """

      plugin_dict = self.__construct_plugins_dictionary(skip_instances=self.proposed_remove_instances)

      plugins_list = []
      for signature, instances in plugin_dict.items():
        plugins_list.append({
          'SIGNATURE': signature,
          'INSTANCES': [instance._get_proposed_config_dictionary(full=True) for instance in instances]
        })

      proposed_pipeline_config = {
        'NAME': self.name,
        'DEFAULT_PLUGIN': False,
        'PLUGINS': plugins_list,
        **self.config,
        **(self.proposed_config or {}),
      }
      return proposed_pipeline_config

    def __register_transactions_for_update(self, session_id: str = None, timeout: float = 0) -> list[Transaction]:
      """
      Register transactions for updating the pipeline and instances configuration. 
      This method is called before sending an update pipeline configuration command to the ratio1 Edge Protocol edge node.

      Parameters
      ----------
      session_id : str, optional
          The session id. A unique id for the session. Defaults to None.

      timeout : int, optional
          The timeout for the transaction. Defaults to 0.

      Returns
      -------
      transactions : list[Transaction]
          The list of transactions generated.
      """
      transactions = []

      # TODO: add different responses for different states of the plugin
      # TODO: maybe this should be introduced as "pre-defined" business plugins, based on a schema
      #       and the user can specify them when creating the pipeline

      for instance in self.lst_plugin_instances:
        if instance._is_tainted():
          transactions.append(self.session._register_transaction(
            session_id=session_id,
            lst_required_responses=instance._get_instance_update_required_responses(),
            timeout=timeout,
            on_success_callback=instance._apply_staged_config,
            # TODO: if the instance was newly added, remove it from the tracked list
            on_failure_callback=instance._discard_staged_config,
          ))
      # end for register to update instances

      for instance in self.proposed_remove_instances:
        transactions.append(self.session._register_transaction(
          session_id=session_id,
          lst_required_responses=instance._get_instance_remove_required_responses(),
          timeout=timeout,
          on_success_callback=lambda: self.__apply_staged_remove_instance(instance),
          on_failure_callback=lambda fail_reason: self.__discard_staged_remove_instance(instance, fail_reason),
        ))
      # end for register to remove instances

      if self.proposed_config is not None:
        required_responses = [
          PipelineOKResponse(self.node_id, self.name),
        ]
        transactions.append(self.session._register_transaction(
          session_id=session_id,
          lst_required_responses=required_responses,
          timeout=timeout,
          on_success_callback=self.__apply_staged_config,
          on_failure_callback=self.__discard_staged_config,
        ))

      return transactions

    def __register_transactions_for_delete(self, session_id: str = None, timeout: float = 0) -> list[Transaction]:
      """
      Register transactions for deleting the pipeline. 
      This method is called before sending a delete pipeline command to the ratio1 Edge Protocol edge node.

      Parameters
      ----------
      session_id : str, optional
          The session id. A unique id for the session. Defaults to None.

      timeout : int, optional
          The timeout for the transaction. Defaults to 0.

      Returns
      -------
      transactions : list[Transaction]
          The list of transactions generated.
      """
      transactions = []

      required_responses = [
        PipelineArchiveResponse(self.node_id, self.name),
      ]
      transactions.append(self.session._register_transaction(
        session_id=session_id,
        lst_required_responses=required_responses,
        timeout=timeout,
        on_success_callback=self.__set_last_operation_successful,
        on_failure_callback=self.__set_last_operation_failed,
      ))

      return transactions

    def __register_transaction_for_pipeline_command(self, session_id: str = None, timeout: float = 0) -> list[Transaction]:
      """
      Register a transaction for a pipeline command. 
      This method is called before sending a pipeline command to the ratio1 Edge Protocol edge node.

      Parameters
      ----------
      session_id : str, optional
          The session id. A unique id for the session. Defaults to None.

      timeout : int, optional
          The timeout for the transaction. Defaults to 0.

      Returns
      -------
      transactions : list[Transaction]
          The list of transactions generated.
      """
      transactions = []

      # TODO: implement
      self.__set_last_operation_successful()

      return transactions

    def __construct_plugins_list(self):
      """
      Construct the plugins list that will be in the pipeline configuration dictionary.

      Returns
      -------
      list
          The plugins list that will be in the pipeline configuration dictionary.
      """
      plugins = []
      dct_signature_instances = {}
      for instance in self.lst_plugin_instances:
        signature = instance.signature
        if instance.signature not in dct_signature_instances:
          dct_signature_instances[instance.signature] = []
        dct_signature_instances[instance.signature].append(instance)
      # end for construct dct_signature_instances

      for signature, instances in dct_signature_instances.items():
        plugins.append({
          'SIGNATURE': signature,
          'INSTANCES': [instance._get_config_dictionary() for instance in instances]
        })
      # end for construct plugins list
      return plugins

    def __construct_plugins_dictionary(self, skip_instances=None):
      """
      Construct the plugins dictionary that will be in the pipeline configuration dictionary.

      Returns
      -------
      dict
          The plugins dictionary that will be in the pipeline configuration dictionary.
      """
      # plugins = []
      skip_instances = skip_instances or []
      dct_signature_instances = {}
      for instance in self.lst_plugin_instances:
        if instance in skip_instances:
          continue
        if instance.signature not in dct_signature_instances:
          dct_signature_instances[instance.signature] = []
        dct_signature_instances[instance.signature].append(instance)
      # end for construct dct_signature_instances

      return dct_signature_instances

    def __send_update_config_to_box(self, session_id=None):
      """
      Send an update pipeline configuration command to the ratio1 Edge Protocol edge node.
      """
      self.session._send_command_update_pipeline_config(
          worker=self.node_addr,
          pipeline_config=self.__get_proposed_pipeline_config(),
          session_id=session_id
      )
      return

    def __batch_update_instances(self, lst_instances, session_id=None):
      """
      Update the configuration of multiple instances at once.
      ```

      Parameters
      ----------
      lst_updates : List[Instance]
          A list of instances.
      """
      lst_updates = []

      for instance in lst_instances:
        lst_updates.append({
          PAYLOAD_DATA.NAME: self.name,
          PAYLOAD_DATA.SIGNATURE: instance.signature,
          PAYLOAD_DATA.INSTANCE_ID: instance.instance_id,
          PAYLOAD_DATA.INSTANCE_CONFIG: instance._get_proposed_config_dictionary(full=False)
        })

      self.session._send_command_batch_update_instance_config(
        worker=self.node_addr,
        lst_updates=lst_updates,
        session_id=session_id
      )

    def __pop_ignored_keys_from_config(self, config):
      """
      Pop the ignored keys from the configuration.

      Parameters
      ----------
      config : dict
          The configuration dictionary.

      Returns
      -------
      dict
          The configuration dictionary without the ignored keys.
          
      Observations
      ------------
      
      Initially the ignored keys were:
        [
          "INITIATOR_ADDR", 
          "INITIATOR_ID", 
          "LAST_UPDATE_TIME", 
          "MODIFIED_BY_ADDR", 
          "MODIFIED_BY_ID"
        ]
      However these keys are essential for the pipeline configuration for app
      monitoring purposes
      """
      ignored_keys = []
      return {k: v for k, v in config.items() if k not in ignored_keys}

    def __get_instance_object(self, signature, instance_id):
      """
      Get the instance object by signature and instance id.

      Parameters
      ----------
      signature : str
          The signature of the plugin.
      instance_id : str
          The name of the instance.

      Returns
      -------
      Instance
          The instance object.
      """
      for instance in self.lst_plugin_instances:
        if instance.signature == signature and instance.instance_id == instance_id:
          return instance
      return None

    def __set_last_operation_successful(self):
      """
      Set the last operation successful.
      """
      self.__was_last_operation_successful = True
      return

    def __set_last_operation_failed(self, fail_reason):
      """
      Set the last operation failed.
      """
      self.__was_last_operation_successful = False
      return

    @staticmethod
    def __custom_exec_on_data(self, instance_id, on_data_callback, data):
      """
      Handle the data received from a custom execution instance. This method is called by the Session object when a message is received from a custom execution instance.

      Parameters
      ----------
      instance_id : str
          The name of the instance that sent the message.
      on_data_callback : Callable[[Pipeline, dict, dict], None]
          The callback that handles the message. The first dict is the payload, and the second dict is the entire message.
      data : dict | Payload
          The payload of the message.
      """
      # TODO: use formatter for this message
      # TODO: expose the other fields from data
      exec_data = None

      exec_data = data.get('EXEC_RESULT', data.get('EXEC_INFO'))
      exec_error = data.get('EXEC_ERRORS', 'no keyword error')

      if exec_error is not None:
        self.P("Error received from <CUSTOM_EXEC_01:{}>: {}".format(instance_id, exec_error), color="r", verbosity=1)
      if exec_data is not None:
        on_data_callback(self, exec_data, data)
      return

    def __apply_staged_remove_instance(self, instance: Instance):
      """
      Remove an instance from the pipeline.

      Parameters
      ----------
      instance : Instance
          The instance to be removed.
      """
      instance.config = None
      try:
        self.__staged_remove_instances.remove(instance)
      except:
        self.P("Attempted to remove instance <{}:{}>, but it was not found in the staged remove list. "
               "Most likely the instance deletion used `with_confirmation=False`".format(
                   instance.signature, instance.instance_id), color="r")
      return

    def __discard_staged_remove_instance(self, instance: Instance, fail_reason: str):
      """
      Discard the removal of an instance from the pipeline.

      Parameters
      ----------
      instance : Instance
          The instance to be removed.
      """

      self.P(
        f"Discarding staged removal of instance <{instance.signature}:{instance.instance_id}>. Reason: {fail_reason}", color="r")

      try:
        self.__staged_remove_instances.remove(instance)
      except:
        self.P("Attempted to remove instance <{}:{}>, but it was not found in the staged remove list. "
               "Most likely the instance deletion used `with_confirmation=False`".format(
                   instance.signature, instance.instance_id), color="r")

      self.lst_plugin_instances.append(instance)
      return

    def __apply_staged_config(self, verbose=False):
      """
      Apply the staged configuration to the pipeline.
      """
      if self.__staged_config is None:
        return

      if verbose:
        self.P("Deployed pipeline <{}> on <{}>".format(self.name, self.node_addr), color="g")
      self.__was_last_operation_successful = True

      self.config = {**self.config, **self.__staged_config}
      self.__staged_config = None

      return

    def __apply_staged_instances_config(self, verbose=False):
      """
      Apply the staged configuration to the instances.
      """
      for instance in self.lst_plugin_instances:
        instance._apply_staged_config(verbose=verbose)

      for instance in self.__staged_remove_instances:
        instance.config = None
        self.lst_plugin_instances.remove(instance)

      self.__staged_remove_instances = []
      return

    def __discard_staged_config(self, fail_reason: str):
      """
      Discard the staged configuration for the pipeline.
      """

      self.P(f'Discarding staged configuration for pipeline <{self.name}>. Reason: {fail_reason}', color="r")
      self.__was_last_operation_successful = False

      self.__staged_config = None
      self.__staged_remove_instances = []
      return

    def __stage_proposed_config(self):
      """
      Stage the proposed configuration.
      """
      if self.proposed_config is not None:
        if self.__staged_config is not None:
          raise ValueError(
            "Pipeline configuration has already been staged, waiting for confirmation from Execution Engine")

        self.__staged_config = self.proposed_config
        self.proposed_config = None

      for instance in self.lst_plugin_instances:
        instance._stage_proposed_config()

      self.__staged_remove_instances.extend(self.proposed_remove_instances)
      self.proposed_remove_instances = []

      self.__was_last_operation_successful = None
      return

    def __print_proposed_changes(self):
      """
      Print the proposed changes to the pipeline.
      """

      if self.proposed_config is not None:
        self.P("Proposed changes to pipeline <{}>:".format(self.name), verbosity=1)
        self.P("  - Current config: {}".format(self.config), verbosity=1)
        self.P("  - New pipeline config: {}".format(self.proposed_config), verbosity=1)

      if len(self.proposed_remove_instances) > 0:
        self.P(
          "  - Remove instances: {}".format([instance.instance_id for instance in self.proposed_remove_instances]), verbosity=1)

      for instance in self.lst_plugin_instances:
        if instance._is_tainted():
          self.P("  - Plugin <{}:{}>:".format(instance.signature, instance.instance_id), verbosity=1)
          self.P("    - Current config: {}".format(instance.config), verbosity=1)
          self.P("    - Proposed config: {}".format(instance.proposed_config), verbosity=1)
      return

    def _close(self, timeout=10):
      """
      Close the pipeline.

      Returns
      -------
      list[Transaction]
          The list of transactions generated.
      """
      transactions = self.__register_transactions_for_delete(timeout=timeout)

      self.__was_last_operation_successful = None

      self.session._send_command_archive_pipeline(
        worker=self.node_addr,
        pipeline_name=self.name,
      )

      return transactions

    def _get_base64_code(self, custom_code):
      """
      Get the base64 code.

      Parameters
      ----------
      custom_code : str | callable
          The custom code.

      Returns
      -------
      str
          The base64 code.
      """
      if custom_code is None:
        return None

      if isinstance(custom_code, str):
        # it is a path
        if os.path.exists(custom_code):
          with open(custom_code, "r") as fd:
            plain_code = "".join(fd.readlines())
        # it is a string
        else:
          try:
            method_name = "_DistributedCustomCodePresets__{}".format(custom_code.lower())
            preset_code = getattr(DistributedCustomCodePresets, method_name)
            plain_code = self.get_function_source_code(preset_code)
          except:
            plain_code = custom_code
      elif callable(custom_code):
        # we have a function
        plain_code = self.get_function_source_code(custom_code)
      else:
        raise Exception("custom_code is not a string or a callable")
      # endif get plain code

      return self.code_to_base64(plain_code, verbose=False)

    
  # Message handling
  if True:
    def _on_data(self, signature, instance_id, data):
      """
      Handle the data received from the ratio1 Edge Protocol edge node. This method is called by the Session object when a message is received from the ratio1 Edge Protocol edge node.
      This method will call all the `on_data` callbacks of the pipeline and the instance that received the message.

      Parameters
      ----------
      signature : str
          The signature of the plugin that sent the message.
      instance_id : str
          The name of the instance that sent the message.
      data : dict | Payload
          The payload of the message.
      """
      # call all self callbacks
      self.Pd(f"Pipeline <{self.name}> received data from <{signature}:{instance_id}>")
      for callback in self.on_data_callbacks:
        callback(self, signature, instance_id, data)

      # call all instance callbacks
      self.__call_instance_on_data_callbacks(signature, instance_id, data)
      return

    def _on_notification(self, signature, instance_id, data):
      """
      Handle the notification received from the ratio1 Edge Protocol edge node. This method is called by the Session object when a notification is received from the ratio1 Edge Protocol edge node.

      Parameters
      ----------
      signature : str
          The signature of the plugin that sent the notification.
      instance_id : str
          The name of the instance that sent the notification.
      data : dict | Payload
          The payload of the notification.
      """
      # call all self callbacks
      for callback in self.on_notification_callbacks:
        callback(self, data)

      # call all instance callbacks
      self.__call_instance_on_notification_callbacks(signature, instance_id, data)
      return

    def _add_on_data_callback(self, callback):
      """
      Add a new callback to the list of callbacks that handle the data received from the pipeline.

      Parameters
      ----------
      callback : Callable[[Pipeline, str, str, dict], None]
          The callback to add
      """
      self.on_data_callbacks.append(callback)
      return

    def _reset_on_data_callback(self):
      """
      Reset the list of callbacks that handle the data received from the pipeline.
      """
      self.on_data_callbacks = []
      return

    def _add_on_notification_callback(self, callback):
      """
      Add a new callback to the list of callbacks that handle the notifications received from the pipeline.

      Parameters
      ----------
      callback : Callable[[Pipeline, dict], None]
          The callback to add
      """
      self.on_notification_callbacks.append(callback)
      return

    def _reset_on_notification_callback(self):
      """
      Reset the list of callbacks that handle the notifications received from the pipeline.
      """
      self.on_notification_callbacks = []
      return

    def __call_instance_on_data_callbacks(self, signature, instance_id, data):
      """
      Call all the `on_data` callbacks of the instance that sent the message.

      Parameters
      ----------
      signature : str
          The signature of the plugin that sent the payload.
      instance_id : str
          The name of the instance that sent the payload.
      data : dict | Payload
          The payload of the payload.
      """
      for instance in self.lst_plugin_instances:
        if instance.signature == signature and instance.instance_id == instance_id:
          instance._on_data(self, data)
      return

    def __call_instance_on_notification_callbacks(self, signature, instance_id, data):
      """
      Call all the `on_notification` callbacks of the instance that sent the notification.

      Parameters
      ----------
      signature : str
          The signature of the plugin that sent the notification.
      instance_id : str
          The name of the instance that sent the notification.
      data : dict | Payload
          The payload of the notification.
      """
      for instance in self.lst_plugin_instances:
        if instance.signature == signature and instance.instance_id == instance_id:
          instance._on_notification(self, data)
      return

  # API
  if True:
    @property
    def was_last_operation_successful(self):
      """
      Return whether the last operation was successful.

      Returns
      -------
      bool
          True if the last operation was successful, False if it failed, None if the ACK has not been received yet
      """
      return self.__was_last_operation_successful

    def create_plugin_instance(
      self, 
      *, 
      signature, 
      instance_id, 
      config={}, 
      on_data=None, 
      on_notification=None, 
      debug=False,
      **kwargs
    ) -> Instance:
      """
      Create a new instance of a desired plugin, with a given configuration. This instance is attached to this pipeline, 
      meaning it processes data from this pipelines data source. Parameters can be passed either in the `config` dict, or as `kwargs`.

      Parameters
      ----------
      signature : str
          The name of the plugin signature. This is the name of the desired overall functionality.
      instance_id : str
          The name of the instance. There can be multiple instances of the same plugin, mostly with different parameters
      config : dict, optional
          parameters used to customize the functionality. One can change the AI engine used for object detection, 
          or finetune alerter parameters to better fit a camera located in a low light environment.
          Defaults to {}
      on_data : Callable[[Pipeline, dict], None], optional
          Callback that handles messages received from this instance. 
          As arguments, it has a reference to this Pipeline object, along with the payload itself.
          Defaults to None
      on_notification : Callable[[Pipeline, dict], None], optional
          Callback that handles notifications received from this instance. 
          As arguments, it has a reference to this Pipeline object, along with the payload itself. 
          Defaults to None

      Returns
      -------
      instance : Instance
          An `Instance` object.

      Raises
      ------
      Exception
          Plugin instance already exists. 
      """
      from ratio1.default.instance import REVERSE_MAP
      
      if isinstance(signature, str):
        str_signature = signature.upper()
        signature = REVERSE_MAP.get(str_signature, str_signature)
      else:
        plugin_template = signature
        str_signature = plugin_template.signature.upper()

      for instance in self.lst_plugin_instances:
        if instance.instance_id == instance_id and instance.signature == str_signature:
          raise Exception("plugin {} with instance {} already exists".format(str_signature, instance_id))

      # create the new instance and add it to the list
      config = {**config, **kwargs}
      instance = self.__init_instance(
        signature=signature, instance_id=instance_id, config=config, 
        on_data=on_data, on_notification=on_notification, is_attached=False,
        debug=debug
      )
      return instance

    def __remove_plugin_instance(self, instance):
      """
      Remove a plugin instance from this pipeline. 

      Parameters
      ----------
      instance : Instance
          The instance to be removed.
      """
      if instance is None:
        raise Exception("The provided instance is None. Please provide a valid instance")

      if instance not in self.lst_plugin_instances:
        raise Exception("plugin  <{}/{}> does not exist on this pipeline".format(instance.signature, instance.instance_id))

      self.lst_plugin_instances.remove(instance)
      return

    def remove_plugin_instance(self, instance):
      """
      Stop a plugin instance from this pipeline. 


      Parameters
      ----------
      instance : Instance
          The instance to be stopped.

      """

      self.__remove_plugin_instance(instance)
      self.proposed_remove_instances.append(instance)
      return

    def create_custom_plugin_instance(self, *, instance_id, custom_code: callable, config={}, on_data=None, on_notification=None, **kwargs) -> Instance:
      """
      Create a new custom execution instance, with a given configuration. This instance is attached to this pipeline, 
      meaning it processes data from this pipelines data source. The code used for the custom instance must be provided
      either as a string, or as a path to a file. Parameters can be passed either in the `config` dict, or as kwargs.
      The custom plugin instance will run periodically. If one desires to execute a custom code only once, use `wait_exec`.

      Parameters
      ----------
      instance_id : str
          The name of the instance. There can be multiple instances of the same plugin, mostly with different parameters
      custom_code : Callable[[CustomPluginTemplate], Any], optional
          A string containing the entire code, a path to a file containing the code as a string or a function with the code.
          This code will be executed remotely on an ratio1 Edge Protocol edge node. Defaults to None.
      config : dict, optional
          parameters used to customize the functionality. One can change the AI engine used for object detection, 
          or finetune alerter parameters to better fit a camera located in a low light environment.
          Defaults to {}
      on_data : Callable[[Pipeline, dict], None], optional
          Callback that handles messages received from this instance. 
          As arguments, it has a reference to this Pipeline object, along with the payload itself.
          Defaults to None
      on_notification : Callable[[Pipeline, dict], None], optional
          Callback that handles notifications received from this instance. 
          As arguments, it has a reference to this Pipeline object, along with the payload itsel. 
          Defaults to None

      Returns
      -------
      instance : Instance
          An `Instance` object.

      Raises
      ------
      Exception
          The code was not provided.
      Exception
          Plugin instance already exists. 
      """

      b64code = self._get_base64_code(custom_code)

      def callback(pipeline, data): return self.__custom_exec_on_data(pipeline, instance_id, on_data, data)
      callback = callback if on_data is not None else None

      return self.create_plugin_instance(
          signature='CUSTOM_EXEC_01',
          instance_id=instance_id,
          config={
              'CODE': b64code,
              **config
          },
          on_data=callback,
          on_notification=on_notification,
          **kwargs
      )

    # TODO: rename this!!!
    def create_chain_dist_custom_plugin_instance(self,
                                                 *,
                                                 main_node_process_real_time_collected_data: any,
                                                 main_node_finish_condition: any,
                                                 main_node_aggregate_collected_data: any,
                                                 worker_node_code: any,
                                                 nr_remote_worker_nodes: int,
                                                 instance_id=None,
                                                 worker_node_pipeline_config={},
                                                 worker_node_plugin_signature='CUSTOM_EXEC_01',
                                                 worker_node_plugin_config={},
                                                 config={},
                                                 on_data=None,
                                                 on_notification=None,
                                                 **kwargs) -> Instance:
      b64code_process_real_time_collected_data = self._get_base64_code(main_node_process_real_time_collected_data)
      b64code_finish_condition = self._get_base64_code(main_node_finish_condition)
      b64code_aggregate_collected_data = self._get_base64_code(main_node_aggregate_collected_data)
      b64code_remote_node = self._get_base64_code(worker_node_code)

      if instance_id is None:
        instance_id = self.name + "_chain_dist_custom_exec_{}".format(self.log.get_unique_id())

      return self.create_plugin_instance(
          signature='PROCESS_REAL_TIME_COLLECTED_DATA_CUSTOM_EXEC_CHAIN_DIST',
          instance_id=instance_id,
          config={
              'CUSTOM_CODE_PROCESS_REAL_TIME_COLLECTED_DATA': b64code_process_real_time_collected_data,
              'CUSTOM_CODE_FINISH_CONDITION': b64code_finish_condition,
              'CUSTOM_CODE_AGGREGATE_COLLECTED_DATA': b64code_aggregate_collected_data,
              'CUSTOM_CODE_REMOTE_NODE': b64code_remote_node,

              'NR_REMOTE_NODES': nr_remote_worker_nodes,

              'NODE_PIPELINE_CONFIG': {
                'stream_type': "Void",
                **worker_node_pipeline_config
              },
              'NODE_PLUGIN_SIGNATURE': worker_node_plugin_signature,
              'NODE_PLUGIN_CONFIG': {
                **worker_node_plugin_config
              },
              **config
          },
          on_data=on_data,
          on_notification=on_notification,
          **kwargs
      )

    def deploy(
      self, 
      with_confirmation=True, 
      wait_confirmation=True, 
      timeout=10, 
      verbose=False
    ):
      """
      This method is used to deploy the pipeline on the ratio1 Edge Protocol edge node. 
      Here we collect all the proposed configurations and send them to the ratio1 Edge Protocol edge node.
      All proposed configs become staged configs.
      After all responses, apply the staged configs to finish the transaction. 
      """
      # generate a unique session id for this deploy operation
      # this session id will be used to track the transactions

      # step 0: print the proposed changes
      if verbose:
        self.__print_proposed_changes()

      # step 1: register transactions for updates
      transactions = []

      if with_confirmation:
        transactions: list[Transaction] = self.__register_transactions_for_update(timeout=timeout)

      # step 1: send the proposed config to the box
      pipeline_config_changed = self.proposed_config is not None
      have_to_remove_instances = len(self.proposed_remove_instances) > 0
      have_new_instances = any([instance._is_tainted() and len(instance.config) == 0
                                for instance in self.lst_plugin_instances])
      if pipeline_config_changed or have_to_remove_instances or have_new_instances:
        # updated pipeline config or deleted instances
        self.__send_update_config_to_box()
      elif any([instance._is_tainted() for instance in self.lst_plugin_instances]):
        # updated instances only
        tainted_instances = [instance for instance in self.lst_plugin_instances if instance._is_tainted()]
        self.__batch_update_instances(tainted_instances)
      else:
        return

      # step 3: stage the proposed config
      self.__stage_proposed_config()

      # step 3: wait for the box to respond
      if with_confirmation and wait_confirmation:
        self.session.wait_for_transactions(transactions)

      # step 4: apply the staged config
      if not with_confirmation:
        self.__apply_staged_config(verbose=verbose)
        self.__apply_staged_instances_config(verbose=verbose)

      # TODO: must add "deployed" message to proper location
      self.P("Pipeline <{}> deploy initiated...".format(self.name))

      if with_confirmation and not wait_confirmation:
        return transactions
      return

    def wait_exec(self, *, custom_code: callable, instance_config={}, timeout=10):
      """
      Create a new REST-like custom execution instance, with a given configuration. This instance is attached to this pipeline, 
      meaning it processes data from this pipelines data source. The code used for the custom instance must be provided either as a string, or as a path to a file. Parameters can be passed either in the config dict, or as kwargs.
      The REST-like custom plugin instance will execute only once. If one desires to execute a custom code periodically, use `create_custom_plugin_instance`.

      Parameters
      ----------
      custom_code : Callable[[CustomPluginTemplate], Any], optional
          A string containing the entire code, a path to a file containing the code as a string or a function with the code.
          This code will be executed remotely on an ratio1 Edge Protocol edge node. Defaults to None.
      config : dict, optional
          parameters used to customize the functionality, by default {}

      Returns
      -------
      Tuple[Any, Any]
          a tuple containing the result of the execution and the error, if any. 
          If the execution completed successfully, the `error` is None, and the `result` is the returned value of the custom code.

      Raises
      ------
      Exception
          The code was not provided.
      Exception
          Plugin instance already exists. 
      """

      b64code = self._get_base64_code(custom_code)

      finished = False
      result = None
      error = None

      def on_data(pipeline, data):
        nonlocal finished
        nonlocal result
        nonlocal error

        if 'REST_EXECUTION_RESULT' in data and 'REST_EXECUTION_ERROR' in data:
          result = data['REST_EXECUTION_RESULT']
          error = data['REST_EXECUTION_ERROR']
          finished = True
        return

      instance_id = self.name + "_rest_custom_exec_synchronous_" + self.log.get_unique_id()
      instance_config = {
          'REQUEST': {
              'DATA': {
                  'CODE': b64code,
              },
              'TIMESTAMP': self.log.time_to_str()
          },
          'RESULT_KEY': 'REST_EXECUTION_RESULT',
          'ERROR_KEY': 'REST_EXECUTION_ERROR',
          **instance_config
      }

      prop_config = self.__get_proposed_pipeline_config()
      if prop_config['TYPE'] == 'Void':
        instance_config['ALLOW_EMPTY_INPUTS'] = True
        instance_config['RUN_WITHOUT_IMAGE'] = True

      self.create_plugin_instance(
          signature='REST_CUSTOM_EXEC_01',
          instance_id=instance_id,
          config=instance_config,
          on_data=on_data
      )

      self.deploy()

      start_time = time()
      while not finished and time() - start_time < timeout:
        sleep(0.1)

      return result, error

    def close(self, wait_confirmation=True, timeout=10):
      """
      Close the pipeline, stopping all the instances associated with it.
      """

      transactions = self._close(timeout=timeout)

      if wait_confirmation:
        self.session.wait_for_transactions(transactions)
      else:
        return transactions
      return

    def P(self, *args, **kwargs):
      """
      Print info to stdout.
      """
      return self.log.P(*args, **kwargs)

    def D(self, *args, **kwargs):
      """
      Call the `Logger.D` method.
      If using the default Logger, this call will print debug info to stdout if `silent` is set to `False`.
      The logger object is passed from the Session object to the Pipeline object when creating
      it with `create_pipeline` or `attach_to_pipeline`.
      """
      return self.session.D(*args, **kwargs)

    def attach_to_plugin_instance(self, signature, instance_id, on_data=None, on_notification=None) -> Instance:
      """
      Attach to an existing instance on this pipeline. 
      This method is useful when one wishes to attach an 
      `on_data` and `on_notification` callbacks to said instance.

      Parameters
      ----------
      signature : str
          name of the plugin signature.
      instance_id : str
          name of the instance.
      on_data : Callable[[Pipeline, dict], None], optional
          Callback that handles messages received from this instance. As arguments, it has a reference to this Pipeline object, along with the payload itself.
          Defaults to None.
      on_notification : Callable[[Pipeline, dict], None], optional
          Callback that handles notifications received from this instance. As arguments, it has a reference to this Pipeline object, along with the payload itself.
          Defaults to None.

      Returns
      -------
      instance : Instance
          An `Instance` object.

      Raises
      ------
      Exception
          the pipeline does not contain plugins with a given signature.
      Exception
          The pipeline does not contain the desired instance.
      """
      from ratio1.default.instance import REVERSE_MAP
      # search for the instance in the list
      plugin_template = None
      if isinstance(signature, str):
        str_signature = signature.upper()
        signature = REVERSE_MAP.get(str_signature, str_signature)
      else:
        plugin_template = signature
        str_signature = plugin_template.signature.upper()
      found_instance = None
      for instance in self.lst_plugin_instances:
        if instance.instance_id == instance_id and instance.signature == str_signature.upper():
          found_instance = instance
          break

      if found_instance is None:
        raise Exception(f"Unable to attach to instance. Instance <{str_signature}/{instance_id}> does not exist")

      # add the callbacks to the session
      if on_data is not None:
        found_instance._add_on_data_callback(on_data)

      if on_notification is not None:
        found_instance._add_on_notification_callback(on_notification)

      if plugin_template is not None:
        found_instance.convert_to_specialized_class(plugin_template)

      return found_instance

    def attach_to_custom_plugin_instance(self, instance_id, on_data=None, on_notification=None) -> Instance:
      """
      Attach to an existing custom execution instance on this pipeline. 
      This method is useful when one wishes to attach an 
      `on_data` and `on_notification` callbacks to said instance.

      Parameters
      ----------
      instance_id : str
          name of the instance.
      on_data : Callable[[Pipeline, str, str, dict], None], optional
          Callback that handles messages received from this instance. As arguments, it has a reference to this Pipeline object, along with the payload itself.
          Defaults to None.
      on_notification : Callable[[Pipeline, dict], None], optional
          Callback that handles notifications received from this instance. As arguments, it has a reference to this Pipeline object, along with the payload itself.
          Defaults to None.

      Returns
      -------
      str
          An identifier for this instance, useful for stopping an instance.

      Raises
      ------
      Exception
          the pipeline does not contain any custom plugin.
      Exception
          The pipeline does not contain the desired instance.
      """

      def callback(pipeline, data): return self.__custom_exec_on_data(pipeline, instance_id, on_data, data)
      callback = callback if on_data is not None else None

      return self.attach_to_plugin_instance("CUSTOM_EXEC_01", instance_id, callback, on_notification)

    def detach_from_instance(self, instance: Instance):
      # search for the instance in the list
      if instance is None:
        raise Exception("The provided instance is None. Please provide a valid instance")

      instance._reset_on_data_callback()
      instance._reset_on_notification_callback()
      return

    def update_acquisition_parameters(self, config={}, **kwargs):
      """
      Update the acquisition parameters of this pipeline.
      Parameters can be passed either in the `config` dict, or as `kwargs`.

      Parameters
      ----------
      config : dict, optional
          The new configuration of the acquisition source, by default {}
      """
      if self.__staged_config is not None:
        raise ValueError("Pipeline configuration has already been staged, waiting for confirmation from Execution Engine")

      if self.proposed_config is None:
        self.proposed_config = {}

      self.proposed_config = {**self.proposed_config, **config, **{k.upper(): v for k, v in kwargs.items()}}
      self.proposed_config = self.__pop_ignored_keys_from_config(self.proposed_config)

      for k, v in self.config.items():
        if k in self.proposed_config:
          if self.proposed_config[k] == v:
            del self.proposed_config[k]

      if len(self.proposed_config) == 0:
        self.proposed_config = None

      return

    def send_pipeline_command(self, command, payload=None, command_params=None, wait_confirmation=True, timeout=10) -> list[Transaction]:
      """
      Send a pipeline command to the ratio1 Edge Protocol edge node.
      This command can block until the command is confirmed by the ratio1 Edge Protocol edge node.

      Example:
      --------
      ```python
      pipeline.send_pipeline_command('START', wait_confirmation=True)

      transactions_p1 = pipeline1.send_pipeline_command('START', wait_confirmation=False)
      transactions_p2 = pipeline2.send_pipeline_command('START', wait_confirmation=False)
      # wait for both commands to be confirmed, but after both commands are sent
      session.wait_for_transactions(transactions_p1 + transactions_p2)
      ```

      Parameters
      ----------
      command : str
          The name of the command.
      payload : dict, optional
          The payload of the command, by default {}
      command_params : dict, optional
          The parameters of the command, by default {}
      wait_confirmation : bool, optional
          Whether to wait for the confirmation of the command, by default False
      timeout : int, optional
          The timeout for the transaction, by default 10    

      Returns
      -------
      list[Transaction] | None
          The list of transactions generated, or None if `wait_confirmation` is False.
      """
      transactions = self.__register_transaction_for_pipeline_command(timeout=timeout)

      self.__was_last_operation_successful = None

      self.session._send_command_pipeline_command(
        worker=self.node_addr,
        pipeline_name=self.name,
        command=command,
        payload=payload,
        command_params=command_params,
      )

      if wait_confirmation:
        self.session.wait_for_transactions(transactions)
      else:
        return transactions
      return

    def create_or_attach_to_plugin_instance(self, *, signature, instance_id, config={}, on_data=None, on_notification=None, **kwargs) -> Instance:
      """
      Create a new instance of a desired plugin, with a given configuration, or attach to an existing instance.

      Parameters
      ----------
      signature : str
          The name of the plugin signature. This is the name of the desired overall functionality.
      instance_id : str
          The name of the instance. There can be multiple instances of the same plugin, mostly with different parameters
      config : dict, optional
          parameters used to customize the functionality. One can change the AI engine used for object detection, 
          or finetune alerter parameters to better fit a camera located in a low light environment.
          Defaults to {}
      on_data : Callable[[Pipeline, dict], None], optional
          Callback that handles messages received from this instance. 
          As arguments, it has a reference to this Pipeline object, along with the payload itself.
          Defaults to None
      on_notification : Callable[[Pipeline, dict], None], optional
          Callback that handles notifications received from this instance. 
          As arguments, it has a reference to this Pipeline object, along with the payload itself. 
          Defaults to None

      Returns
      -------
      instance : Instance
          An `Instance` object.
      """
      try:
        instance = self.attach_to_plugin_instance(signature, instance_id, on_data, on_notification)
        instance.update_instance_config(config, **kwargs)
      except Exception as e:
        instance = self.create_plugin_instance(
          signature=signature,
          instance_id=instance_id,
          config=config,
          on_data=on_data,
          on_notification=on_notification,
          **kwargs
        )
      return instance

    def create_or_attach_to_custom_plugin_instance(self, *, instance_id, custom_code, config={}, on_data=None, on_notification=None, **kwargs) -> Instance:
      """
      Create a new instance of a desired plugin, with a given configuration, or attach to an existing instance. 

      Parameters
      ----------
      signature : str
          The name of the plugin signature. This is the name of the desired overall functionality.
      instance_id : str
          The name of the instance. There can be multiple instances of the same plugin, mostly with different parameters
      config : dict, optional
          parameters used to customize the functionality. One can change the AI engine used for object detection, 
          or finetune alerter parameters to better fit a camera located in a low light environment.
          Defaults to {}
      on_data : Callable[[Pipeline, dict], None], optional
          Callback that handles messages received from this instance. 
          As arguments, it has a reference to this Pipeline object, along with the payload itself.
          Defaults to None
      on_notification : Callable[[Pipeline, dict], None], optional
          Callback that handles notifications received from this instance. 
          As arguments, it has a reference to this Pipeline object, along with the payload itself. 
          Defaults to None

      Returns
      -------
      instance : Instance
          An `Instance` object.
      """

      try:
        instance = self.attach_to_custom_plugin_instance(instance_id, on_data, on_notification)
        instance.update_instance_config(config, **kwargs)
      except:
        instance = self.create_custom_plugin_instance(
          instance_id=instance_id,
          custom_code=custom_code,
          config=config,
          on_data=on_data,
          on_notification=on_notification,
          **kwargs
        )
      return instance
    
    
    def _sync_configuration_with_remote(self, config={}, plugins_statuses : list = None):
      """
      Given a configuration, update the pipeline configuration and the 
      instances configuration.
            
      """
      config.pop('NAME', None)
      config.pop('TYPE', None)
      plugins = config.pop('PLUGINS', {})

      self.config = {**self.config, **config}
      
      self.__update_plugins_statuses_data(plugins_statuses)

      active_plugins = []
      for dct_signature_instances in plugins:
        signature = dct_signature_instances['SIGNATURE']
        instances = dct_signature_instances['INSTANCES']
        for dct_instance in instances:
          instance_id = dct_instance.pop('INSTANCE_ID')
          active_plugins.append((signature, instance_id))
          instance_object = self.__get_instance_object(signature, instance_id)
          if instance_object is None:
            instance_object = self.__init_instance(signature, instance_id, dct_instance, None, None, is_attached=True) # here the plugin status is updated if data is available
          else:
            instance_object._sync_configuration_with_remote(dct_instance) 
          # next we update the plugin status from known plugins statuses
          self._update_plugin_status(instance_object)
        # end for dct_instance
      # end for dct_signature_instances

      for instance in self.lst_plugin_instances:
        if (instance.signature, instance.instance_id) not in active_plugins:
          self.__remove_plugin_instance(instance)
      # end for instance
      return

    def update_full_configuration(self, config={}):
      """
      Update the full configuration of this pipeline.
      Parameters are passed in the `config` dict.
      We do not support kwargs yet because it makes it difficult to check priority of dictionary, merging values, etc.

      Parameters
      ----------
      config : dict, optional
          The new configuration of the pipeline, by default {}
      """
      if self.__staged_config is not None:
        raise ValueError("Pipeline configuration has already been staged, waiting for confirmation from Execution Engine")

      # pop the illegal to modify keys
      config.pop('NAME', None)
      config.pop('TYPE', None)
      plugins = config.pop('PLUGINS', None)

      self.update_acquisition_parameters(config)

      if plugins is None:
        return

      new_plugins = []
      for dct_signature_instances in plugins:
        signature = dct_signature_instances['SIGNATURE']
        instances = dct_signature_instances['INSTANCES']
        for dct_instance in instances:
          instance_id = dct_instance.pop('INSTANCE_ID')
          new_plugins.append((signature, instance_id))
          instance_object = self.__get_instance_object(signature, instance_id)

          if instance_object is None:
            self.create_plugin_instance(signature=signature, instance_id=instance_id, config=dct_instance)
          else:
            instance_object.update_instance_config(dct_instance)
        # end for dct_instance
      # end for dct_signature_instances

      # now check if we have to remove any instances
      for instance in self.lst_plugin_instances:
        if (instance.signature, instance.instance_id) not in new_plugins:
          self.remove_plugin_instance(instance)
      # end for instance
      return

    @property
    def node_alias(self):
      """
      Return the node alias of the pipeline.
      """
      return self.session.get_node_alias(self.node_addr)

    @property
    def node_id(self):
      """
      Method alias for `node_alias`
      """
      return self.node_alias

    @property
    def node_name(self):
      """
      Method alias for `node_alias`
      """
      return self.node_alias

    def get_full_config(self):
      """
      Get the full configuration of the pipeline.
      Returns
      -------

      """
      dct_signature_instances = {}
      for instance in self.lst_plugin_instances:
        signature = instance.signature
        if signature not in dct_signature_instances:
          dct_signature_instances[signature] = []
        dct_signature_instances[signature].append(instance)
      # endfor instances
      return {
        'NAME': self.name,
        'PLUGINS': [
          {
            'SIGNATURE': signature,
            'INSTANCES': [
              {
                'INSTANCE_ID': instance.instance_id,
                **instance.config
              }
              for instance in instances
            ]
          }
          for signature, instances in dct_signature_instances.items()
        ],
        **self.config
      }



"""
ex04_custom_code_on_one_remote__example_2.py
--------------------------------------------

This is a simple example of how to use the ratio1 SDK.

In this example, we connect to the network, choose a node and
    deploy a plugin with custom code that will run in real time.
    
For this example, we search for prime numbers using more than one thread.

The difference from previous example is that we define the custom code that will run in parallel
  on the client size, and we send it to the remote side as base64 code.
"""
from ratio1 import Session, CustomPluginTemplate


def remote_code_is_prime(plugin: CustomPluginTemplate, data: int):
  """
  Code that will be executed remote and in parallel, on multiple threads.

  Parameters
  ----------
  plugin : CustomPluginTemplate
      The plugin instance. It will be replaced with the plugin instance object on the remote side. 
  data : int
      The data that will be processed by the remote code.
      In our case, it will be a random number, but it can be anything

  Returns
  -------
  bool
      The result of this code.
      In our case, True if the number is prime, False otherwise
  """
  n = data

  if n <= 1:
    return False
  for i in range(2, int(n**0.5) + 1):
    if n % i == 0:
      return False
  return True


def plugin_custom_code(plugin: CustomPluginTemplate):
  """
  The custom code that will be executed on the main thread.

  Parameters
  ----------
  plugin : CustomPluginTemplate
      The plugin instance. It will be replaced with the plugin instance object on the remote side.

  Returns
  -------
  list
      The result of the custom code.
      In our case, the list of prime numbers found in the random numbers generated.
  """
  base64_custom_method = plugin.cfg_threading_custom_code

  random_numbers = plugin.np.random.randint(0, 100, 10)

  are_primes = plugin.threadapi_base64_code_map(base64_custom_method, random_numbers)

  return [n for n, is_prime in zip(random_numbers, are_primes) if is_prime]


def on_data(pipeline, result, full_payload):
  print("Data received: ", result)


if __name__ == "__main__":
  session = Session()

  session.wait_for_any_node()

  node = session.get_active_nodes()[0]

  pipeline = session.create_or_attach_to_pipeline(
    node=node,
    name="run_threading_api",
    data_source="Void"
  )

  pipeline.create_or_attach_to_custom_plugin_instance(
    instance_id="run_threading_api_01",
    custom_code=plugin_custom_code,
    process_delay=10,

    # we convert the code that will be executed on multiple threads to base64
    # to send it to the remote side
    threading_custom_code=pipeline.method_to_base64(remote_code_is_prime),
    on_data=on_data
  )

  pipeline.deploy()


  # Observation:
  #   next code is not mandatory - it is used to keep the session open and cleanup the resources
  #   in production, you would not need this code as the script can close after the pipeline will be sent
  session.run(
    wait=60, # wait for the user to stop the execution or a given time
    close_pipelines=True # when the user stops the execution, the remote edge-node pipelines will be closed
  )


from ratio1 import Instance, Payload, Pipeline, Session


if __name__ == '__main__':

  session: Session = Session()

  # Please, fill in below your node's address
  # this code assumes the node have "allowed" the SDK to deploy the pipeline
  node_address = ''


  session.wait_for_node(node_address)  # we wait for the node to be ready
  pipeline: Pipeline = session.create_pipeline(
    node=node_address,
    name='home_camera_security_pipeline',
    data_source="video_stream_cv2",
    # Fill in below the rtsp URL for your camera and define cap_resolution to camera's FPS + 1
    url='',
    cap_resolution=16,
  )

  instance: Instance = pipeline.create_plugin_instance(
    signature='HOME_SAFETY',
    instance_id='inst01',
    debuggin_mode=True,
  )

  pipeline.deploy()

  session.wait(
    # seconds=99999999,  # we wait the session for 60 seconds
    close_pipelines=False,  # we close the pipelines after the session
    close_session=False,  # we close the session after the session
    wait_close=False
  )
  session.P("Main thread exiting...")



import os
import json

from time import time
from ratio1.utils.config import log_with_color
from ratio1.const import SESSION_CT, COMMANDS, BASE_CT
from ratio1._ver import __VER__ as version

from pandas import DataFrame
from datetime import datetime


def _get_netstats(
  silent=True,
  online_only=False, 
  allowed_only=False, 
  supervisor=None,
  alias_filter=None,
  supervisors_only=False,
  return_session=False,
  eth=False,
  all_info=False,
  wait_for_node=None
):
  t1 = time()
  from ratio1 import Session
  sess = Session(silent=silent)
  found = None
  if wait_for_node:
    sess.P("Waiting for node '{}' to appear...".format(wait_for_node), color='y')
    found = sess.wait_for_node(wait_for_node, timeout=30)
    if not found:
      sess.P("Node '{}' not found.".format(wait_for_node), color='r')

  dct_info = sess.get_network_known_nodes(
    online_only=online_only, allowed_only=allowed_only, supervisor=supervisor,
    supervisors_only=supervisors_only,
    alias_filter=alias_filter,
    min_supervisors=1,
    eth=eth,
    all_info=all_info, 
  )
  df = dct_info[SESSION_CT.NETSTATS_REPORT]
  supervisor = dct_info[SESSION_CT.NETSTATS_REPORTER]
  super_alias = dct_info[SESSION_CT.NETSTATS_REPORTER_ALIAS]
  nr_supers = dct_info[SESSION_CT.NETSTATS_NR_SUPERVISORS]
  _elapsed = dct_info[SESSION_CT.NETSTATS_ELAPSED] # computed on call
  elapsed = time() - t1 # elapsed=_elapsed
  if return_session:
    return df, supervisor, super_alias, nr_supers, elapsed, sess  
  return df, supervisor, super_alias, nr_supers, elapsed



def get_nodes(args):
  """
  This function is used to get the information about the nodes and it will perform the following:
  
  1. Create a Session object.
  2. Wait for the first net mon message via Session and show progress. 
  3. Wait for the second net mon message via Session and show progress.  
  4. Get the active nodes union via Session and display the nodes marking those peered vs non-peered.
  """
  supervisor_addr = args.supervisor
  alias_filter = args.alias
  online = args.online
  online = True # always online, flag deprecated
  wide = args.wide
  if args.verbose:
    log_with_color(f"Getting nodes from supervisor <{supervisor_addr}>...", color='b')

  res = _get_netstats(
    silent=not args.verbose,
    online_only=online or args.peered,
    allowed_only=args.peered,
    supervisor=supervisor_addr,
    alias_filter=alias_filter,
    eth=args.eth,
    all_info=wide,
    return_session=True,
  )
  df, supervisor, super_alias, nr_supers, elapsed, sess = res
  if args.online:
    FILTERED = ['State']
    df = df[[c for c in df.columns if c not in FILTERED]]

  prefix = "Online n" if (online or args.peered) else "N"
  # network = os.environ.get(BASE_CT.dAuth.DAUTH_NET_ENV_KEY, BASE_CT.dAuth.DAUTH_SDK_NET_DEFAULT)
  network = sess.bc_engine.evm_network
  addr = sess.bc_engine.address
  if supervisor == "ERROR":
    log_with_color(f"No supervisors or no comms available in {elapsed:.1f}s. Please check your settings.", color='r')
  else:
    log_with_color(f"Ratio1 client v{version}: {addr} \n", color='b')
    log_with_color(
      "{}odes on '{}' reported by <{}> '{}' in {:.1f}s ({} supervisors seen):".format(
      prefix, network, supervisor, super_alias, elapsed, nr_supers), 
      color='b'
    )
    import pandas as pd
    pd.set_option('display.float_format', '{:.4f}'.format)
    log_with_color(f"{df}\n")    
  return
  
  
def get_supervisors(args):
  """
  This function is used to get the information about the supervisors.
  """
  if args.verbose:
    log_with_color("Getting supervisors...", color='b')

  res = _get_netstats(
    silent=not args.verbose,
    online_only=True,
    supervisors_only=True,
    return_session=True,
  )
  df, supervisor, super_alias, nr_supers, elapsed, sess = res
  FILTERED = ['Oracle', 'State']
  df = df[[c for c in df.columns if c not in FILTERED]]
  
  if supervisor == "ERROR":
    log_with_color(f"No supervisors or no comms available in {elapsed:.1f}s. Please check your settings.", color='r')
  else:
    log_with_color(
      "Supervisors on '{}' reported by <{}> '{}' in {:.1f}s".format(
      sess.bc_engine.evm_network, supervisor, super_alias, elapsed), 
      color='b'
    )
    log_with_color(f"{df}")
  return


def get_apps(args):
  """
  Shows the apps running on a given node, if the client is allowed on that node.
  Parameters
  ----------
  args : argparse.Namespace
      Arguments passed to the function.

  """
  verbose = args.verbose
  node = args.node
  show_full = args.full
  as_json = args.json
  owner = args.owner

  # 1. Init session
  from ratio1 import Session
  sess = Session(
    silent=not verbose
  )
  
  res = sess.get_nodes_apps(
    node=node, owner=owner, show_full=show_full, 
    as_json=as_json, as_df=not as_json
  )
  if res is not None:
    network = sess.bc_engine.evm_network
    node_alias = sess.get_node_alias(node) if node else None
    if as_json:
      log_with_color(json.dumps(res, indent=2))
    else:
      df_apps = res
      if df_apps.shape[0] == 0:
        log_with_color(
          "No user apps found on node <{}> '{}' of network '{}'".format(
            node, node_alias, network            
          ), 
          color='r'
        )
        return
      # remove Node column
      if node is not None and owner is None:
        df_apps.drop(columns=['Node'], inplace=True)
      
      if node is None and owner is not None:
        df_apps.drop(columns=['Owner'], inplace=True)
      
      if node is not None:
        last_seen = sess.get_last_seen_time(node)
        last_seen_str = datetime.fromtimestamp(last_seen).strftime('%Y-%m-%d %H:%M:%S') if last_seen else None
        is_online = sess.check_node_online(node)    
        node_status = 'Online' if is_online else 'Offline'
      else:
        last_seen_str = "N/A"
        node_status = "N/A"
      #end if node
      if node == None:
        node = "[All available]"
      by_owner = f" by owner <{owner}>" if owner else ""    
      log_with_color(f"Ratio1 client v{version}:\n", color='b')
      log_with_color(
        "Apps on <{}> ({}) [Status: {}| Last seen: {}]{}:".format(
          node, network, node_status, last_seen_str, by_owner
        ), 
        color='b'
      )
      log_with_color(f"{df_apps}\n")
    #end if as_json
  #end if res is not None
  return

def _send_command_to_node(args, command, ignore_not_found=False):
  node = args.node
  silent = not args.verbose   

  
  t1 = time()
  df, _, _, _, _, sess = _get_netstats(
    silent=silent, online_only=True, return_session=True, all_info=True,
    wait_for_node=node
  )
  
  peered = None
  selection = (df.Alias == node) | (df.Address == node)
  if 'ETH Address' in df.columns:
    selection = selection | (df['ETH Address'] == node)
  found = selection.any()
  node_addr = None
  alias = None
  df_found =  df[selection]
  if found:
    alias = df_found.Alias.values[0]
    peered = df_found.Peered.values[0]
    node_addr = df_found.Address.values[0]   
    log_with_color(f"{df_found}")
  else:
    log_with_color("Node '{}' <{}> not found in network (toal {} nodes, {} peered).".format(
      node, node_addr, df.shape[0], df.Peered.sum()), color='r'
    )
    node_addr = node
    
  if not peered:
    if found:
      log_with_color(f"Node '{node}' <{node_addr}> is not peered.", color='r')
    else:
      log_with_color(f"Node '{node}' <{node_addr}> may not accept this command.", color='r')
    
  # TODO: currently this is based on node alias, but we should be based on node address
  #       and maybe even node alias
  if (found and peered) or ignore_not_found:
    if found and peered:
      log_with_color(f"Sending '{command}' to node '{alias}' <{node_addr}>", color='b')
    else:
      log_with_color(f"Sending blind '{command}' to node '{alias}' <{node_addr}>", color='b')      
    if command == COMMANDS.RESTART:
      sess._send_command_restart_node(node_addr)
    elif command == COMMANDS.STOP:
      sess._send_command_stop_node(node_addr)
    else:
      log_with_color(f"Command '{command}' not supported.", color='r')
      return
  elapsed = time() - t1  
  return  

def restart_node(args):
  """
  This function is used to restart the node.
  
  Parameters
  ----------
  args : argparse.Namespace
      Arguments passed to the function.
  """
  node = args.node
  log_with_color(f"Attempting to restart node <{node}>", color='b')
  _send_command_to_node(args, COMMANDS.RESTART, ignore_not_found=True)
  return


def shutdown_node(args):
  """
  This function is used to shutdown the node.
  
  Parameters
  ----------
  args : argparse.Namespace
      Arguments passed to the function.
  """
  node = args.node
  log_with_color(f"Attempting to shutdown node <{node}>", color='b')
  _send_command_to_node(args, COMMANDS.STOP, ignore_not_found=True)
  return




import os
from time import sleep, time

from ratio1 import Payload, Pipeline, Session, load_dotenv


boxes = {}

custom_worker_code = """
_result=None
payload = plugin.dataapi_struct_data()
if payload is not None:
  plugin.P("Received data")

  payload_box_name = payload.get('EE_ID', payload.get('SB_ID'))
  payload_active_plugins = payload.get('ACTIVE_PLUGINS')

  if payload_box_name is not None and payload_active_plugins is not None:


    inst = []

    for active_plugin in payload_active_plugins:
      inst.append((active_plugin['STREAM_ID'], active_plugin['SIGNATURE'], active_plugin['INSTANCE_ID']))

    _result = {
      'BOX' : payload_box_name,
      'INSTANCES' : inst
    }
  else:
    plugin.P(payload)
    plugin.P(payload['EE_EVENT_TYPE'])
    plugin.P(payload_box_name)
    plugin.P(payload_active_plugins)
"""


def instance_on_data(pipeline: Pipeline, data: Payload):
  global boxes
  data['INSTANCES'].sort()
  boxes[data['BOX']] = (data['INSTANCES'], time())
  return


def pipeline_on_notification(pipeline, notification: dict):
  pipeline.P(
      "Received specific notification for pipeline {}".format(pipeline.name))
  return


def sess_on_hb(sess, node, hb):
  act_plug = hb['ACTIVE_PLUGINS']
  for plug in act_plug:
    sess.P((node, plug['STREAM_ID'], plug['SIGNATURE'], plug['INSTANCE_ID']))


load_dotenv()

listener_params = {
  'HOST': os.getenv('AIXP_HOSTNAME'),
  'PORT': int(os.getenv('AIXP_PORT')),
  'USER': os.getenv('AIXP_USERNAME'),
  'PASS': os.getenv('AIXP_PASSWORD'),
  'TOPIC': 'lummetry/ctrl',
}

node_id = 'node_id'
sess = Session(
  on_heartbeat=sess_on_hb
)

pipeline = sess.create_pipeline(
    node=node_id,
    name='test_mqtt',
    data_source='IotQueueListener',
    config={
        'STREAM_CONFIG_METADATA': listener_params,
        "RECONNECTABLE": True,
    },
    plugins=None,
    on_notification=pipeline_on_notification
)

# now start a ciclic process
pipeline.create_custom_plugin_instance(
    instance_id='inst01',
    # plain_code_path="./custom_exec_scripts/custom_exec_tutorial.txt", # you can provide it as a file
    plain_code=custom_worker_code,
    on_data=instance_on_data,
    process_delay=2
)

pipeline.deploy()


def generate_net_map():
  global boxes
  net_map_msgs = ["Printing network map"]
  for box, (instances, last_time) in boxes.items():
    if last_time - time() > 30:
      continue
    net_map_msgs.append(box + ":")
    for i in range(len(instances)):
      (pipeline, signature, instance) = instances[i]

      (o_pipeline, o_signature) = (None, None)
      if i != 0:
        (o_pipeline, o_signature, _) = instances[i - 1]

      same_pipeline = pipeline == o_pipeline
      same_signature = signature == o_signature

      if not same_pipeline:
        net_map_msgs.append(" --- " + pipeline + ":")
      if not same_pipeline or not same_signature:
        net_map_msgs.append("    --- " + signature + ":")
      net_map_msgs.append("        --- " + instance)

    net_map_msgs.append("")
  return "\n".join(net_map_msgs)


try:
  while True:
    sleep(10)

    # pipeline.P(generate_net_map(), color='m')
except KeyboardInterrupt:
  pipeline.P("CTRL+C detected. Closing example..", color='r')


sess.close(close_pipelines=True)



#!/usr/bin/env python3
from web3 import Web3

# Connect to Base mainnet RPC (chain ID 8453)
base_rpc = "https://base-mainnet.public.blastapi.io"
w3 = Web3(Web3.HTTPProvider(base_rpc))

# Uniswap V2 R1/USDC pair contract address
pair_address = Web3.to_checksum_address("0x0feC06fd2C2bd4066c7302c08950aBaA2E4AB1d3")

# Minimal ABI for Uniswap V2 pair (token0, token1, getReserves, totalSupply, balanceOf, and Transfer event)
pair_abi = [
    {"constant": True, "inputs": [], "name": "token0", "outputs": [{"name": "", "type": "address"}], "type": "function"},
    {"constant": True, "inputs": [], "name": "token1", "outputs": [{"name": "", "type": "address"}], "type": "function"},
    {"constant": True, "inputs": [], "name": "getReserves", "outputs": [
        {"name": "reserve0", "type": "uint112"},
        {"name": "reserve1", "type": "uint112"},
        {"name": "blockTimestampLast", "type": "uint32"}], "type": "function"},
    {"constant": True, "inputs": [], "name": "totalSupply", "outputs": [{"name": "", "type": "uint"}], "type": "function"},
    {"constant": True, "inputs": [{"name": "owner", "type": "address"}], "name": "balanceOf", "outputs": [{"name": "", "type": "uint"}], "type": "function"},
    {"anonymous": False, "inputs": [
        {"indexed": True, "name": "from", "type": "address"},
        {"indexed": True, "name": "to", "type": "address"},
        {"indexed": False, "name": "value", "type": "uint256"}],
     "name": "Transfer", "type": "event"}
]
pair_contract = w3.eth.contract(address=pair_address, abi=pair_abi)

# Fetch token addresses and pool data
token0 = pair_contract.functions.token0().call()       # Address of R1 or USDC
token1 = pair_contract.functions.token1().call()       # Address of the other token
reserve0, reserve1, _ = pair_contract.functions.getReserves().call()
total_supply = pair_contract.functions.totalSupply().call()

# Determine which reserve is R1 vs USDC by checking token symbols (optional)
erc20_abi = [{"constant": True, "inputs": [], "name": "symbol", "outputs": [{"name": "", "type": "string"}], "type": "function"}]
sym0 = w3.eth.contract(address=token0, abi=erc20_abi).functions.symbol().call().upper()
sym1 = w3.eth.contract(address=token1, abi=erc20_abi).functions.symbol().call().upper()
if sym0 == 'R1':
    r1_reserve, usdc_reserve = reserve0, reserve1
elif sym1 == 'R1':
    r1_reserve, usdc_reserve = reserve1, reserve0
else:
    # Fallback assumption if symbol check fails
    r1_reserve, usdc_reserve = reserve0, reserve1

# Prepare Transfer event signature hash for filtering logs
transfer_topic = Web3.keccak(text="Transfer(address,address,uint256)").hex()

# Fetch all Transfer events from block 0 to latest for this pair contract
logs = w3.eth.get_logs({
    "fromBlock": 0,
    "toBlock": "latest",
    "address": pair_address,
    "topics": [transfer_topic]
})

# Collect all addresses that appear in Transfer logs
all_addresses = set()
for log in logs:
    # Topics[1] = from, Topics[2] = to (indexed fields)
    from_addr = "0x" + log['topics'][1].hex()[-40:]
    to_addr   = "0x" + log['topics'][2].hex()[-40:]
    # Convert to checksum addresses
    from_addr = Web3.to_checksum_address(from_addr)
    to_addr   = Web3.to_checksum_address(to_addr)
    # Add addresses if not zero address
    if from_addr != "0x0000000000000000000000000000000000000000":
        all_addresses.add(from_addr)
    if to_addr != "0x0000000000000000000000000000000000000000":
        all_addresses.add(to_addr)

# Filter for EOAs with positive LP balance
holders = {}
for addr in all_addresses:
    balance = pair_contract.functions.balanceOf(addr).call()
    if balance > 0:
        code = w3.eth.get_code(addr)
        # EOA check: get_code returns empty (0x) for EOAs:contentReference[oaicite:12]{index=12}
        if code == b'' or code == '0x' or code == b'0x' or code == b'0x0':
            holders[addr] = balance

# Compute and print each holder's share of R1 and USDC
for holder, bal in holders.items():
    share = bal / total_supply
    r1_amount = int(share * r1_reserve)
    usdc_amount = int(share * usdc_reserve)
    print(f"{holder}: R1={r1_amount}, USDC={usdc_amount}")



#!/usr/bin/env python3

"""
ex19_deeploy_example.py
---------------------

This tutorial demonstrates how to interact with the Deeploy API using the ratio1 SDK.
It shows how to:
- Build and sign messages for Deeploy API requests
- Send authenticated requests to the Deeploy API
- Handle responses

The example includes a CLI client that can be used to:
- Create pipelines
- Delete pipelines
- Get apps

Example Usage:
-------------
1. Get list of apps:
   ```bash
   python3 ex19_deeploy_example.py --private-key path/to/private-key.pem --endpoint get_apps > res.json
   ```

2. Create a pipeline:
   ```bash
   python3 ex19_deeploy_example.py --private-key path/to/private-key.pem --request path/to/request.json --endpoint create_pipeline
   ```
   
   Example request.json for pipeline creation:
   ```json
   {
     "request": {
       "app_alias": "app_deployed_from_tutorials",
       "plugin_signature": "PLUGIN_SIGNATURE",
       "target_nodes": [
         "0xai_target_node_1"
       ],
       "target_nodes_count": 0
       "pipeline_input_type": "void"
     }
   }
   ```

3. Delete a pipeline:
   ```bash
   python3 ex19_deeploy_example.py --private-key path/to/private-key.pem --request path/to/request.json --endpoint delete_pipeline
   ```

   Example request.json for pipeline deletion:
   ```json
   {
     "request": {
       "app_id": "target_app_name_id_returned_by_get_apps_or_create_pipeline",
       "target_nodes": [
         "0xai_target_node_1"
       ]
     }
   }
   ```

4. Send app command:
    ```bash
    python3 ex19_deeploy_example.py --private-key path/to/private-key.pem --request path/to/request.json --endpoint send_app_command
    ```

    Example request.json for sending app command:
    ```json
    {
      "request": {
        "app_id": "target_app_name_id_returned_by_get_apps_or_create_pipeline",
        "app_command": "RESTART"
      }
    }
    ```

5. Send instance command:
    ```bash
    python3 ex19_deeploy_example.py --private-key path/to/private-key.pem --request path/to/request.json --endpoint send_instance_command
    ```
    Example request.json for sending instance command:
    ```json
    {
      "request": {
        "app_id": "target_app_name_id_returned_by_get_apps_or_create_pipeline",
        "target_nodes": [
          "0xai_target_node_1"
        ],
        "plugin_signature": "PLUGIN_SIGNATURE",
        "instance_id": "PLUGIN_INSTANCE_ID",
        "instance_command": "RESTART"
      }
    }
    ```

Note: The private key file should be in PEM format (typically with .pem extension) and contain your Ethereum private key.
The request JSON file should contain the appropriate request data for the endpoint you're calling.



```python

from ratio1 import Session

if __name__ == '__main__':
  # we do not setup any node as we will not use direct SDK deployment but rather the Deeploy API
  sess = Session()

  launch_result = sess.deeploy_launch_container_app(
    image="tvitalii/flask-docker-app:latest",
    name="ratio1_simple_container_webapp",
    port=5000,  
    # signer_key_path="../../path/to/private-key.pem",
    # signer_key_password=None,  # if your private key has a password, set it here
    signer_key_string=str_metamask_key,  # if you want to use a private key string instead of a file
    target_nodes=["0xai_target_node_1"],  # replace with your target node address
    # target_nodes_count=0,  # if you want to deploy to all nodes, set this to 0
  )

  # no neeed for further `sess.deploy()` as the `deeploy_*` methods handle the deployment automatically
  # now we interpret the launch_result and extract app-id, etc
  # ...

  # if all ok sleep for a while to allow app testing (say 60 seconds)

  # finally use deeploy close

  close_result = sess.deeploy_close(
    ... # use `launch_result` to get the necessary parameters
  )
  
  # log the result
```

"""

import json
import os
import time
import argparse

import requests

from ratio1 import Logger
from ratio1.bc import DefaultBlockEngine
from ratio1.const.base import BCct

API_BASE_URL = "https://devnet-deeploy-api.ratio1.ai"


if __name__ == "__main__":
  parser = argparse.ArgumentParser(description='Deeploy CLI client')
  parser.add_argument('--private-key', type=str, required=True, help='Path to PEM private key file')
  parser.add_argument('--key-password', type=str, required=False, help='Private key password (if PK has any).',
                      default=None)
  parser.add_argument('--request', type=str, required=False, help='Path to request JSON file',
                      default=None)
  parser.add_argument('--endpoint', type=str, default='create_pipeline',
                      choices=['create_pipeline', 'delete_pipeline', 'get_apps', 'send_app_command', 'send_instance_command'],
                      help='API endpoint to call')

  args = parser.parse_args()
  logger = Logger("DEEPLOY", base_folder=".", app_folder="ex_19_deeploy_example")

  private_key_path = args.private_key
  private_key_password = args.key_password
  endpoint = args.endpoint

  # Check if PK exists
  if not os.path.isfile(args.private_key):
    print("Error: Private key file does not exist.")
    exit(1)

  # Read request data if provided or use default empty request
  if args.request:
    with open(args.request, 'r') as f:
      request_data = json.load(f)
  else:
    request_data = {'request': {}}

  try:
    # Set the nonce for the request
    nonce = f"0x{int(time.time() * 1000):x}"
    request_data['request']['nonce'] = nonce

    # Create a block engine instance with the private key
    block_engine = DefaultBlockEngine(
      log=logger,
      name="default",
      config={
        BCct.K_PEM_FILE: f"../../{private_key_path}",
        BCct.K_PASSWORD: private_key_password,
      }
    )

    # Sign the payload using eth_sign_payload
    signature = block_engine.eth_sign_payload(
      payload=request_data['request'],
      indent=1,
      no_hash=True,
      message_prefix="Please sign this message for Deeploy: "
    )

    # Send request
    response = requests.post(f"{API_BASE_URL}/{endpoint}", json=request_data)
    response = response.json()

    print(json.dumps(response, indent=2))
  except Exception as e:
    print(f"Error: {str(e)}")
    exit(1)



import os
import inspect
import importlib
import sys
import traceback

from pkgutil import iter_modules

from copy import deepcopy

from .code_cheker.base import BaseCodeChecker

class _PluginsManagerMixin:

  def __init__(self):
    super(_PluginsManagerMixin, self).__init__()
    self.code_checker = BaseCodeChecker()
    return
  
  @property
  def is_secured(self):
    val = self.log.config_data.get("SECURED", False)
    return val in [True, 'True', 'true', '1', 1]

  def __get_avail_plugins(self, locations, debug=True):
    """
    WARNING: This function is deprecated and should not be used. Now `_get_plugin_by_name` will 
             works without this helper function.
             
             The function has been changed from protected to private and should NOT be used. Was
             left here only for documentation purposes
    """
    if not isinstance(locations, list):
      locations = [locations]    
    names, modules = [], []
    for plugins_location in locations:
      path = plugins_location.replace('.', '/')
      if False:
        # if we use path-based search we will have a big issue with compiled modules!
        path = os.path.join(self.log.code_base_folder, path) # code base can be in another folder !
        if not os.path.isdir(path):
          self.P("*** attemped to search for plugins in a invalid folder: {}".format(path), color='error')
          continue
        all_files = os.listdir(path)
        files = [os.path.splitext(x)[0] for x in all_files if '.py' in x and '__init__' not in x]
      else:
        found_modules = iter_modules([path])
        files = [x.name for x in found_modules]
      if debug:
        self.P("  Found {} in '{}'".format(files, path))
      modules += [plugins_location + '.' + x for x in files]
      names += [x.replace('_', '').lower() for x in files]

    return names, modules

  def get_package_base_path(self, package_name):
    """
    Return the file path of an installed package parent directory.
    
    :param package_name: The name of the installed package.
    :return: The path to the package parent.
    """
    spec = importlib.util.find_spec(package_name)
    if spec is not None and spec.submodule_search_locations:
      return os.path.dirname(spec.submodule_search_locations[0])
    else:
      self.P("Package '{}' not found.".format(package_name), color='r')
    return None

  def _get_plugin_by_name(self, lst_plugins_locations, name, search_in_packages=None, safe=False):
    name = self.log.camel_to_snake(name)

    if search_in_packages is None:
      search_in_packages = []

    total_sub_locations = []
    # First we extract all the sublocations for each package

    for package in search_in_packages:
      package_path = self.get_package_base_path(package)
      if package_path is None:
        continue

      for location in lst_plugins_locations:
        base_location_name = location.split('.')[0]
        # if the location is not in the package we skip it (locations should begin with the package name)
        # TODO: review this please
        if base_location_name != package:
          continue

        root_location = location.replace('.', os.path.sep)
        package_root_location = os.path.join(package_path, root_location)
        sub_locations = self.log.get_all_subfolders(package_root_location, as_package=True)
        sub_locations = [x.replace(package_path.replace(os.path.sep, '.').strip('.'), '').strip('.') for x in sub_locations]
        sub_locations.append(location)
        total_sub_locations += sub_locations
      # end for package locations
    # end for package

    # Then we extract all the sublocations for local files
    for location in lst_plugins_locations:
      root_location = location.replace('.', os.path.sep)
      sub_locations = self.log.get_all_subfolders(root_location, as_package=True)
      total_sub_locations += sub_locations
    # endfor local

    # we remove duplicates
    total_sub_locations = list(set(total_sub_locations))
    lst_plugins_locations = lst_plugins_locations + total_sub_locations
    for loc in lst_plugins_locations:
      if loc.endswith('__pycache__'):
        continue
      candidate = loc + '.' + name
      try:
        found = importlib.util.find_spec(candidate)
        # if found something
        # and that thing can be loaded
        if found is not None and found.loader is not None:
          self.P("    Trying {}: '{}' -> FOUND!".format("[SAFE]" if safe else "[UNSAFE]", candidate))
          return candidate
        else:
          self.P("    Trying {}: '{}' -> NOT found.".format("[SAFE]" if safe else "[UNSAFE]", candidate))
      except ModuleNotFoundError:
        self.P("      Invalid package: '{}'".format(candidate))
    return
  
    
  def _perform_module_safety_check(self, module, safe_imports=None):
    good = True
    msg = ''
    str_code = inspect.getsource(module)
    self.P("  Performing code safety check on module `{}`:".format(module.__name__), color='m')
    errors = self.code_checker.check_code_text(str_code, safe_imports=safe_imports)
    if errors is not None:      
      info = "  ERROR: Unsafe code in {}:\n{}".format(
        module.__name__, '\n'.join(
          ['  *** ' + msg + 'at line(s): {}'.format(v) for msg, v in errors.items()]
          ))
      self.P(info, color='error')
      self._create_notification(
            notif="EXCEPTION", # ct.STATUS_TYPE.STATUS_EXCEPTION, -- TODO: review & change after planning dependency tree
            msg="Unsafe code in `{}`".format(module.__name__),
            info=info,
            modulename=module.__name__,
            autocomplete_info=True,
            printed=True,
          )
      if self.is_secured:
        good = False
        msg = info
      else:
        self.P("   ********* Pluging accepted due to UNSECURED node. **************", color='error')
        good = True
    else:
      self.P("   * * * Module '{}' code check successful * * *".format(module.__name__), color='g')
    # endif bad code
    self.P("  Finished performing code safety check on module `{}`".format(module.__name__), color='m')
    return good, msg

  def _perform_class_safety_check(self, classdef):
    good = True
    msg = ''
    str_code = inspect.getsource(classdef)
    self.P("  Performing class code safety check on class `{}`".format(classdef.__name__), color='m')
    ### TODO: finish code analysis using BaseCodeChecker
    self.P("  Finished class code safety check on class `{}`".format(classdef.__name__), color='m')
    return good, msg

  def _get_module_name_and_class(self, 
                                 locations, 
                                 name,
                                 search_in_packages=None,
                                 suffix=None, 
                                 verbose=1, 
                                 safety_check=False, 
                                 safe_locations=None, 
                                 safe_imports=None,
                                 class_safety_check=False,                                  
                                 ):
    if not isinstance(locations, list):
      locations = [locations]
      

    _class_name, _cls_def, _config_dict = None, None, None
    simple_name = name.replace('_','')

    if suffix is None:
      suffix = ''

    suffix = suffix.replace('_', '')
        
    _safe_module_name = None
    is_safe_plugin = False
    # first search is safe locations always!
    if safe_locations is not None:
      if not isinstance(safe_locations, list):
        safe_locations = [safe_locations]
      if len(safe_locations) > 0:
        _safe_module_name = self._get_plugin_by_name(safe_locations, name, search_in_packages=search_in_packages, safe=True)
        
    if safe_imports is not None and not isinstance(safe_imports, list):
      safe_imports = [safe_imports]
    
    # not a safe module so we search in normal locations
    if _safe_module_name is None:
      # in packages you should have only safe locations
      _user_module_name = self._get_plugin_by_name(locations, name, search_in_packages=None, safe=False)
      _module_name = _user_module_name
    else:
      is_safe_plugin = True 
      _module_name = _safe_module_name
    
    
    if _module_name is None:
      if verbose >= 1:
        self.P("Error with finding plugin '{}' in locations '{}'".format(name, locations), color='r')
      return _module_name, _class_name, _cls_def, _config_dict

    self.P("  Found {} plugin '{}'".format(
      'safe' if is_safe_plugin else 'user', _module_name), 
      color='g' if is_safe_plugin else 'm'
    )
    
    safety_check = False if is_safe_plugin else safety_check
    class_safety_check = False if is_safe_plugin else class_safety_check

    module = None
    try:
      if _module_name in sys.modules:
        del sys.modules[_module_name]
      module = importlib.import_module(_module_name)
      if module is not None and safety_check:
        is_good, msg = self._perform_module_safety_check(module, safe_imports=safe_imports)
        if not is_good:
          err_msg = "CODE SAFETY VIOLATION: {}".format(msg)
          raise ValueError(err_msg)
      classes = inspect.getmembers(module, inspect.isclass)
      for _cls in classes:
        if _cls[0].upper() == simple_name.upper() + suffix.upper():
          _class_name, _cls_def = _cls
      if _class_name is None:
        if verbose >= 1:
          self.P("ERROR: Could not find class match for {} (suffix: {}). Available classes are: {}".format(
            simple_name, suffix, [x[0] for x in classes]
          ), color='r')
      _config_dict_from_file = getattr(module, "_CONFIG", {})
      _version = getattr(module, "__VER__", None) or getattr(module, "__VER__", None)
      _version = _version or "0.0.0"
      # incredibly enough if we do not deepcopy somehow data will be overriden and
      # used in subsequent imports: basically the module remains in the memory
      # and the changed dict will be created at subsequent importlib.import !
      # change below to False to replicate issue !
      if isinstance(_config_dict_from_file, dict) and True: 
        _config_dict = deepcopy(_config_dict_from_file)
      else:
        _config_dict = _config_dict_from_file
      
      _config_dict['MODULE_VERSION'] = _version # added module version if available to the config dict
        
      if _cls_def is not None and class_safety_check:
        is_good, msg = self._perform_class_safety_check(_cls_def)
        if not is_good:
          raise ValueError("Unsafe class code exception: {}".format(msg))     
      _found_location = ".".join(_module_name.split('.')[:-1])
      self.P("  Plugin '{}' loaded and code checked from {}".format(name, _found_location), color='g')
    except:
      str_err = traceback.format_exc()
      msg = "Error preparing {} with module {}:\n{}".format(
        name, _module_name, str_err)
      self.P(msg, color='error')
      module, _class_name, _cls_def, _config_dict = None, None, None, None

    return module, _class_name, _cls_def, _config_dict



"""
ex06_custom_code_on_multiple_remotes__example_1.py
---------------------------------

This is a simple example of how to use the ratio1 SDK in order to run custom code on 
multiple remote nodes.

"""
from ratio1 import Session, CustomPluginTemplate, Pipeline
from ratio1 import DistributedCustomCodePresets as Presets


def local_brute_force_prime_number_generator():
  from concurrent.futures import ThreadPoolExecutor
  import numpy as np

  def is_prime(n):
    if n <= 1:
      return False
    for i in range(2, int(np.sqrt(n)) + 1):
      if n % i == 0:
        return False
    return True

  thread_pool = ThreadPoolExecutor(max_workers=4)

  random_numbers = np.random.randint(1, 5000, 20)
  are_primes = list(thread_pool.map(is_prime, random_numbers))

  prime_numbers = []
  for i in range(len(random_numbers)):
    if are_primes[i]:
      prime_numbers.append(random_numbers[i])

  return prime_numbers


def worker_brute_force_prime_number_generator(plugin: CustomPluginTemplate):
  def is_prime(n):
    if n <= 1:
      return False
    for i in range(2, int(plugin.np.sqrt(n)) + 1):
      if n % i == 0:
        return False
    return True

  random_numbers = plugin.np.random.randint(1, 5000, 20)
  are_primes = plugin.threadapi_map(is_prime, random_numbers, n_threads=4)

  prime_numbers = []
  for i in range(len(random_numbers)):
    if are_primes[i]:
      prime_numbers.append(random_numbers[i])

  return prime_numbers


final_result = []


def locally_process_partial_results(pipeline: Pipeline, full_payload):
  global final_result
  progress = full_payload.get('PROGRESS')
  data = full_payload.get('DATA')
  len_data = 0
  if data:
    data.sort()
    len_data = len(data)

  pipeline.P(f"Found: {len_data} -- Primes: {data}\n\n")

  if progress == 100:
    pipeline.P("FINISHED\n\n")
    final_result = data
  return


if __name__ == "__main__":
  s = Session()

  s.wait_for_any_node()

  node = s.get_active_nodes()[0]

  # define the job
  s.create_chain_dist_custom_job(
    node=node,
    main_node_process_real_time_collected_data=Presets.PROCESS_REAL_TIME_COLLECTED_DATA__KEEP_UNIQUES_IN_AGGREGATED_COLLECTED_DATA,
    main_node_finish_condition=Presets.FINISH_CONDITION___AGGREGATED_DATA_MORE_THAN_X,
    main_node_finish_condition_kwargs={
      "X": 100
    },
    main_node_aggregate_collected_data=Presets.AGGREGATE_COLLECTED_DATA___AGGREGATE_COLLECTED_DATA,
    nr_remote_worker_nodes=2,

    worker_node_code=worker_brute_force_prime_number_generator,
    worker_node_plugin_config={
      "MAX_INT": 5000,
    },

    on_data=locally_process_partial_results,
    # if True, we wait until our node confirms that it received the job
    deploy=True
  )


  # Observation:
  #   next code is not mandatory - it is used to keep the session open and cleanup the resources
  #   in production, you would not need this code as the script can close after the pipeline will be sent
  # process incoming messages until the finish condition is met
  s.run(wait=lambda: len(final_result) == 0, close_pipelines=True, close_session=True)

  s.P("Final result has {} observations".format(len(final_result)), color='g')



from .misc import COLORS, WEEKDAYS_SHORT, SESSION_CT
from . import comms as COMMS
from . import base as BASE_CT
from . import payload as PAYLOAD_CT
from .formatter import FORMATTER_DATA
from .payload import STATUS_TYPE, PAYLOAD_DATA, COMMANDS, NOTIFICATION_CODES, NET_CONFIG
from .base import CONFIG_STREAM, BIZ_PLUGIN_DATA, PLUGIN_INFO, BLOCKCHAIN_CONFIG
from . import heartbeat as HB
from .environment import ENVIRONMENT
from .apps import PLUGIN_SIGNATURES, DEFAULT_PIPELINES
